# test_locator -- Test nxt.locator module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import Mock, call, patch

import pytest

import nxt.locator


def make_backend_mock():
    m = Mock(spec_set=("get_backend",))
    backend = Mock(spec_set=("find",))
    backend.find.return_value = []
    m.get_backend.return_value = backend
    return m


@pytest.fixture
def mbackend_usb():
    return make_backend_mock()


@pytest.fixture
def mbackend_bluetooth():
    return make_backend_mock()


@pytest.fixture
def mbackend_devfile():
    return make_backend_mock()


@pytest.fixture
def mbackend_socket():
    return make_backend_mock()


@pytest.fixture(autouse=True)
def mimportlib(mbackend_usb, mbackend_bluetooth, mbackend_devfile, mbackend_socket):
    def import_module(name):
        if name == "nxt.backend.usb":
            return mbackend_usb
        if name == "nxt.backend.bluetooth":
            return mbackend_bluetooth
        if name == "nxt.backend.devfile":
            return mbackend_devfile
        if name == "nxt.backend.socket":
            return mbackend_socket
        raise ImportError("no such module")

    with patch("nxt.locator.importlib") as m:
        m.import_module.side_effect = import_module
        yield m


@pytest.fixture(autouse=True)
def mconfigparser():
    parser = Mock(spec_set=("read", "__contains__", "__getitem__"))
    parser.read.return_value = ["some", "files"]
    parser.__contains__ = Mock(return_value=True)
    parser.__getitem__ = Mock(return_value=dict())
    with patch("nxt.locator.configparser") as m:
        m.ConfigParser.return_value = parser
        m.DEFAULTSECT = "DEFAULT"
        yield m


def test_find_no_found():
    with pytest.raises(nxt.locator.BrickNotFoundError):
        nxt.locator.find()


def test_find_first(mbackend_usb, mbackend_bluetooth, mbrick, mbrick2):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    mbackend_bluetooth.get_backend().find.return_value = [mbrick2]
    assert nxt.locator.find() is mbrick


def test_find_first_second_backend(mbackend_usb, mbackend_bluetooth, mbrick):
    mbackend_usb.get_backend().find.return_value = []
    mbackend_bluetooth.get_backend().find.return_value = [mbrick]
    assert nxt.locator.find() is mbrick


def test_find_by_name(mbackend_usb, mbrick, mbrick2):
    mbackend_usb.get_backend().find.return_value = [mbrick, mbrick2]
    mbrick.get_device_info.return_value = "NXT", None, None, None
    mbrick2.get_device_info.return_value = "NXT2", None, None, None
    assert nxt.locator.find(name="NXT2") is mbrick2


def test_find_by_host(mbackend_usb, mbrick, mbrick2):
    mbackend_usb.get_backend().find.return_value = [mbrick, mbrick2]
    mbrick.get_device_info.return_value = "NXT", "00:16:53:00:00:01", None, None
    mbrick2.get_device_info.return_value = "NXT2", "00:16:53:00:00:02", None, None
    assert nxt.locator.find(host="00:16:53:00:00:02") is mbrick2


def test_find_by_custom(mbackend_usb, mbrick, mbrick2):
    mbackend_usb.get_backend().find.return_value = [mbrick, mbrick2]
    assert nxt.locator.find(custom_match=lambda b: b is mbrick2) is mbrick2


def test_find_all(mbackend_usb, mbackend_bluetooth, mbrick, mbrick2):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    mbackend_bluetooth.get_backend().find.return_value = [mbrick2]
    bricks = list(nxt.locator.find(find_all=True))
    assert len(bricks) == 2
    assert bricks[0] == mbrick
    assert bricks[1] == mbrick2


def test_find_filename(mbackend_devfile, mbrick):
    mbackend_devfile.get_backend().find.return_value = [mbrick]
    assert nxt.locator.find(filename="/dev/rfcomm0") is mbrick


def test_find_socket(mbackend_socket, mbrick):
    mbackend_socket.get_backend().find.return_value = [mbrick]
    assert nxt.locator.find(server_host="localhost") is mbrick
    assert nxt.locator.find(server_port=2727) is mbrick


def test_find_bad_backend():
    with pytest.raises(ValueError):
        assert nxt.locator.find(backends=["bad!"])


def test_find_backend_not_found():
    with pytest.raises(ImportError):
        assert nxt.locator.find(backends=["unknown"])


def test_find_no_config(mbackend_usb, mconfigparser, mbrick):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    assert nxt.locator.find(config=None) is mbrick
    assert not mconfigparser.ConfigParser.called


def test_find_no_config_section(mbackend_usb, mconfigparser, mbrick):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    parser = mconfigparser.ConfigParser.return_value
    parser.__contains__.return_value = False
    assert nxt.locator.find() is mbrick
    assert parser.__contains__.mock_calls == [
        call("default"),
    ]
    assert parser.__getitem__.mock_calls == [
        call("DEFAULT"),
    ]


def test_find_config_backends(
    mbackend_usb,
    mbackend_bluetooth,
    mbackend_devfile,
    mbackend_socket,
    mconfigparser,
    mbrick,
):
    mbackend_socket.get_backend().find.return_value = [mbrick]
    parser = mconfigparser.ConfigParser.return_value
    parser.__getitem__.return_value = dict(backends="devfile socket")
    assert nxt.locator.find() is mbrick
    assert not mbackend_usb.get_backend().find.called
    assert not mbackend_bluetooth.get_backend().find.called
    assert mbackend_devfile.get_backend().find.called
    assert mbackend_socket.get_backend().find.called


def test_find_config_backends_override(
    mbackend_usb, mbackend_socket, mconfigparser, mbrick
):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    parser = mconfigparser.ConfigParser.return_value
    parser.__getitem__.return_value = dict(backends="socket")
    assert nxt.locator.find(backends=["usb"]) is mbrick
    assert mbackend_usb.get_backend().find.called
    assert not mbackend_socket.get_backend().find.called


def test_find_config(mbackend_usb, mconfigparser, mbrick):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    mbrick.get_device_info.return_value = "NXT", "00:16:53:00:00:01", None, None
    parser = mconfigparser.ConfigParser.return_value
    parser.__getitem__.return_value = dict(name="NXT", host="00:16:53:00:00:01")
    assert nxt.locator.find() is mbrick
    assert mbackend_usb.get_backend().find.mock_calls == [
        call(name="NXT", host="00:16:53:00:00:01"),
    ]


def test_find_config_override(mbackend_usb, mconfigparser, mbrick):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    mbrick.get_device_info.return_value = "NXT2", "00:16:53:00:00:02", None, None
    parser = mconfigparser.ConfigParser.return_value
    parser.__getitem__.return_value = dict(name="NXT", host="00:16:53:00:00:01")
    assert nxt.locator.find(name="NXT2") is mbrick
    assert mbackend_usb.get_backend().find.mock_calls == [
        call(name="NXT2", host=None),
    ]


def test_find_config_filters(mbackend_usb, mconfigparser, mbrick):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    parser = mconfigparser.ConfigParser.return_value
    parser.__getitem__.return_value = dict(test="testing")
    assert nxt.locator.find() is mbrick
    assert mbackend_usb.get_backend().find.mock_calls == [
        call(name=None, host=None, test="testing"),
    ]


def test_find_config_filenames(mbackend_usb, mconfigparser, mbrick):
    mbackend_usb.get_backend().find.return_value = [mbrick]
    parser = mconfigparser.ConfigParser.return_value
    assert nxt.locator.find(config_filenames=["some", "files"]) is mbrick
    assert parser.read.mock_calls == [
        call(["some", "files"]),
    ]



# test_motcont -- Test nxt.motcont module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import call

import pytest

import nxt.error
import nxt.motcont
import nxt.motor


@pytest.fixture
def mc(mbrick):
    return nxt.motcont.MotCont(mbrick)


def msg(x):
    """Remove spaces and encode to bytes."""
    return x.replace(" ", "").encode("ascii")


def test_cmd(mbrick, mtime, mc):
    mc.cmd(nxt.motor.Port.B, -100, 1000, speedreg=1, smoothstart=0, brake=0)
    # Using the same port yield a delay.
    mc.cmd(nxt.motor.Port.B, 10, 0, speedreg=0, smoothstart=1, brake=1)
    assert mbrick.mock_calls == [
        call.message_write(1, msg("1 1 200 001000 2")),
        call.message_write(1, msg("1 1 010 000000 5")),
    ]
    assert mtime.sleep.mock_calls == [
        call(0.015),
    ]


def test_cmd_nosleep(mbrick, mtime, mc):
    mc.cmd(nxt.motor.Port.B, -100, 1000, speedreg=1, smoothstart=0, brake=0)
    # Using a different port yield no delay.
    mc.cmd(nxt.motor.Port.C, 10, 0, speedreg=0, smoothstart=1, brake=1)
    assert mbrick.mock_calls == [
        call.message_write(1, msg("1 1 200 001000 2")),
        call.message_write(1, msg("1 2 010 000000 5")),
    ]
    assert mtime.sleep.mock_calls == []


def test_cmd_twomotors(mbrick, mtime, mc):
    mc.cmd(nxt.motor.Port.B, -100, 1000, speedreg=1, smoothstart=0, brake=0)
    # When using two motors, there should be a delay as the same motor is used again.
    mc.cmd(
        (nxt.motor.Port.B, nxt.motor.Port.C), 10, 0, speedreg=0, smoothstart=1, brake=1
    )
    assert mbrick.mock_calls == [
        call.message_write(1, msg("1 1 200 001000 2")),
        call.message_write(1, msg("1 5 010 000000 5")),
    ]
    assert mtime.sleep.mock_calls == [
        call(0.015),
    ]


def test_cmd_threemotors(mbrick, mtime, mc):
    with pytest.raises(ValueError):
        mc.cmd(
            (nxt.motor.Port.A, nxt.motor.Port.B, nxt.motor.Port.C),
            -100,
            1000,
            speedreg=1,
            smoothstart=0,
            brake=0,
        )


def test_reset_tacho(mbrick, mc):
    mc.reset_tacho(nxt.motor.Port.B)
    assert mbrick.mock_calls == [call.message_write(1, msg("2 1"))]


def test_is_ready(mbrick, mtime, mc):
    mbrick.message_read.side_effect = [(1, msg("1 1")), (1, msg("2 0"))]
    ready = mc.is_ready(nxt.motor.Port.B)
    not_ready = mc.is_ready(nxt.motor.Port.C)
    assert mtime.sleep.called
    assert mbrick.mock_calls == [
        call.message_write(1, msg("3 1")),
        call.message_read(0, 1, 1),
        call.message_write(1, msg("3 2")),
        call.message_read(0, 1, 1),
    ]
    assert ready is True
    assert not_ready is False


def test_is_ready_error(mbrick, mc):
    mbrick.message_read.return_value = (1, msg("0 1"))
    with pytest.raises(nxt.error.ProtocolError):
        mc.is_ready(nxt.motor.Port.B)


def test_set_output_state(mbrick, mc):
    mc.set_output_state(nxt.motor.Port.C, -100, 1000, speedreg=1)
    mc.set_output_state(nxt.motor.Port.C, 10, 0, speedreg=0)
    assert mbrick.mock_calls == [
        call.message_write(1, msg("4 2 200 001000 1")),
        call.message_write(1, msg("4 2 010 000000 0")),
    ]


def test_start(mbrick, mc):
    mc.start()
    assert mbrick.mock_calls == [
        call.stop_program(),
        call.start_program("MotorControl22.rxe"),
    ]


def test_start_not_running(mbrick, mc):
    mbrick.stop_program.side_effect = [nxt.error.DirectProtocolError("")]
    mc.start()
    assert mbrick.mock_calls == [
        call.stop_program(),
        call.start_program("MotorControl22.rxe"),
    ]


def test_stop(mbrick):
    mc = nxt.motcont.MotCont(mbrick)
    mc.stop()
    assert mbrick.mock_calls == [
        call.stop_program(),
    ]



# nxt.sensor module -- Classes to read LEGO Mindstorms NXT sensors
# Copyright (C) 2006,2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, Paulo Vieira, rhn
# Copyright (C) 2010  Marcus Wanner
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import enum

__all__ = ["Port", "Type", "Mode", "Sensor"]


class Port(enum.Enum):
    """Input port identifier.

    The prefix is needed because a Python identifier can not start with a digit.
    """

    S1 = 0
    """Sensor port 1."""

    S2 = 1
    """Sensor port 2."""

    S3 = 2
    """Sensor port 3."""

    S4 = 3
    """Sensor port 4."""


class Type(enum.Enum):
    """Sensor type."""

    NO_SENSOR = 0
    """No sensor is connected."""

    SWITCH = 1
    """Touch sensor."""

    TEMPERATURE = 2
    """RCX temperature sensor."""

    REFLECTION = 3
    """RCX light sensor."""

    ANGLE = 4
    """RCX rotation sensor."""

    LIGHT_ACTIVE = 5
    """Light sensor with light active."""

    LIGHT_INACTIVE = 6
    """Light sensor with light off."""

    SOUND_DB = 7
    """Sound sensor (unadjusted)."""

    SOUND_DBA = 8
    """Sound sensor (adjusted)."""

    CUSTOM = 9
    """Custom sensor (unused)."""

    LOW_SPEED = 10
    """Low speed digital sensor."""

    LOW_SPEED_9V = 11
    """Low speed digital sensor with 9V supply voltage."""

    HIGH_SPEED = 12
    """High speed sensor."""

    COLOR_FULL = 13
    """NXT color sensor in full color mode (color sensor mode)."""

    COLOR_RED = 14
    """NXT color sensor with red light on (light sensor mode)."""

    COLOR_GREEN = 15
    """NXT color sensor with green light on (light sensor mode)."""

    COLOR_BLUE = 16
    """NXT color sensor in with blue light on (light sensor mode)."""

    COLOR_NONE = 17
    """NXT color sensor in with light off (light sensor mode)."""

    COLOR_EXIT = 18
    """NXT color sensor internal state."""


class Mode(enum.Enum):
    """Sensor mode."""

    RAW = 0x00
    """Raw value, from 0 to 1023."""

    BOOL = 0x20
    """Boolean value, 0 or 1."""

    EDGE = 0x40
    """Count number of transitions."""

    PULSE = 0x60
    """Count number of pulse."""

    PERCENT = 0x80
    """Value from 0 to 100."""

    CELSIUS = 0xA0
    """Temperature in degree Celsius."""

    FAHRENHEIT = 0xC0
    """Temperature in degree Fahrenheit."""

    ROTATION = 0xE0
    """RCX rotation sensor mode."""


class Sensor:
    """Sensor base class."""

    def __init__(self, brick, port):
        self._brick = brick
        self._port = port

    def set_input_mode(self, sensor_type, sensor_mode):
        """Set sensor input mode.

        :param Type sensor_type: Sensor type.
        :param Mode sensor_mode: Sensor mode.
        """
        self._brick.set_input_mode(self._port, sensor_type, sensor_mode)



# nxt.backend.devfile module -- Device file backend
# Copyright (C) 2013  Dave Churchill, Marcus Wanner
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import glob
import logging
import platform
import struct
import tty

import nxt.brick

logger = logging.getLogger(__name__)


class DevFileSock:
    """Device file socket connected to a NXT brick."""

    #: Block size.
    bsize = 118

    #: Connection type, used to evaluate latency.
    type = "bluetooth"

    def __init__(self, filename):
        self._filename = filename

    def __str__(self):
        return f"DevFile ({self._filename})"

    def connect(self):
        """Connect to NXT brick.

        :return: Connected brick.
        :rtype: Brick
        """
        logger.info("connecting via %s", self._filename)
        self._device = open(self._filename, "r+b", buffering=0)
        tty.setraw(self._device)
        return nxt.brick.Brick(self)

    def close(self):
        """Close the connection."""
        if self._device is not None:
            logger.info("closing %s connection", self._filename)
            self._device.close()
            self._device = None

    def send(self, data):
        """Send raw data.

        :param bytes data: Data to send.
        """
        data = struct.pack("<H", len(data)) + data
        logger.debug("send: %s", data.hex())
        self._device.write(data)

    def recv(self):
        """Receive raw data.

        :return: Received data.
        :rtype: bytes
        """
        data = self._device.read(2)
        logger.debug("recv: %s", data.hex())
        (plen,) = struct.unpack("<H", data)
        data = self._device.read(plen)
        logger.debug("recv: %s", data.hex())
        return data


class Backend:
    """Device file backend.

    This uses a device file present on Linux or macOS /dev file system, which allows to
    connect to the NXT brick without needing a Bluetooth python package.

    You only need to use this backend if the :mod:`~nxt.backend.bluetooth` backend is
    not working for you.

    On Linux, you need to pair the NXT brick, then you can use the rfcomm tool::

        sudo rfcomm bind 0 00:16:53:01:02:03

    Where ``00:16:53:01:02:03`` is the Bluetooth address of your NXT brick. This will
    create a ``/dev/rfcomm0`` device file which can be used to communicate with the NXT
    brick.

    On macOS, you need to pair the NXT brick, then open Bluetooth preferences, select
    the NXT brick, click “Edit serial ports”. It should show “NXT-DevB-1”. If not, add
    a serial port using:

    - Port name: NXT-DevB-1
    - Device service: Dev B
    - Port type: RS-232

    This should create a ``/dev/tty.NXT-DevB-1`` device file which can be used to
    communicate with the NXT brick.
    """

    def find(self, name=None, filename=None, **kwargs):
        """Find bricks connected using Bluetooth using device file.

        :param name: Brick name (example: ``"NXT"``).
        :type name: str or None
        :param filename: Device file name (example: ``"/dev/rfcomm0"``).
        :type filename: str or None
        :param kwargs: Other parameters are ignored.
        :return: Iterator over all found bricks.
        :rtype: Iterator[Brick]
        """
        if filename:
            matches = [filename]
        else:
            system = platform.system()
            if system == "Linux":
                matches = glob.glob("/dev/rfcomm*")
            elif system == "Darwin":
                if name:
                    matches = glob.glob("/dev/*%s*" % name)
                else:
                    matches = glob.glob("/dev/*-DevB*")
            else:
                matches = []
        for match in matches:
            sock = DevFileSock(match)
            try:
                brick = sock.connect()
            except OSError:
                logger.exception("failed to connect to device %s", sock)
            else:
                yield brick


def get_backend():
    """Get an instance of the device file backend.

    :return: Device file backend.
    :rtype: Backend
    """
    return Backend()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "NXT-Python"
copyright = "2021, Nicolas Schodet"
author = "Nicolas Schodet"

# The full version, including alpha/beta/rc tags
release = "3.1.0"
version = release


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "3.4"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]

nitpicky = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_logo = "../logo.svg"
html_favicon = "favicon.ico"
html_copy_source = False

# -- Options for autodoc -----------------------------------------------------

autodoc_member_order = "bysource"



# nxt.sensor.analog module -- submodule for use with analog sensors
# Copyright (C) 2006,2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, Paulo Vieira, rhn
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import nxt.sensor


class RawReading:
    """A object holding the raw sensor values for a sensor."""

    def __init__(self, values):
        (
            self.port,
            self.valid,
            self.calibrated,
            self.sensor_type,
            self.mode,
            self.raw_value,
            self.normalized_value,
            self.scaled_value,
            self.calibrated_value,
        ) = values

    def __repr__(self):
        return str(
            (
                self.port,
                self.valid,
                self.calibrated,
                self.sensor_type,
                self.mode,
                self.raw_value,
                self.normalized_value,
                self.scaled_value,
                self.calibrated_value,
            )
        )


class BaseAnalogSensor(nxt.sensor.Sensor):
    """Object for analog sensors."""

    def get_input_values(self):
        """Get raw sensor readings.

        :return: An object with the read values.
        :rtype: RawReading
        """
        return RawReading(self._brick.get_input_values(self._port))

    def reset_input_scaled_value(self):
        """Reset sensor scaled value.

        This can be used to reset accumulated value for some sensor modes.
        """
        self._brick.reset_input_scaled_value(self._port)



# nxt.brick module -- Classes to represent LEGO Mindstorms NXT bricks
# Copyright (C) 2006  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, rhn
# Copyright (C) 2010  rhn, Marcus Wanner, zonedabone
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import io
import threading
import time

import nxt.error
import nxt.motor
import nxt.sensor
import nxt.sensor.digital
from nxt.telegram import Opcode, Telegram

__all__ = ["Brick"]


class RawFileReader(io.RawIOBase):
    """Implement RawIOBase for reading a file on the NXT brick."""

    def __init__(self, brick, name):
        self._brick = brick
        self._handle, self._remaining = brick.file_open_read(name)

    def close(self):
        if not self.closed:
            super().close()
            self._brick.file_close(self._handle)

    def readable(self):
        return True

    def readinto(self, b):
        rsize = min(self._brick._sock.bsize, self._remaining, len(b))
        if rsize == 0:
            return 0
        _, data = self._brick.file_read(self._handle, rsize)
        size = len(data)
        self._remaining -= size
        b[0:size] = data
        return size


class RawFileWriter(io.RawIOBase):
    """Implement RawIOBase for writing a file on the NXT brick."""

    def __init__(self, brick, name, size):
        self._brick = brick
        self._handle = brick.file_open_write(name, size)
        self._remaining = size

    def close(self):
        if not self.closed:
            super().close()
            self._brick.file_close(self._handle)

    def writable(self):
        return True

    def write(self, b):
        if self.closed:
            raise ValueError("write to closed file")
        if self._remaining == 0:
            raise ValueError("write to a full file")
        wsize = min(self._brick._sock.bsize, self._remaining, len(b))
        _, size = self._brick.file_write(self._handle, bytes(b[:wsize]))
        self._remaining -= size
        return size


class Brick:
    """Object connected to a NXT brick.

    It provides low level access to brick commands and high level access to internal
    brick components (file system, modules...). It can be used to create motor or sensor
    objects.

    Create an instance with :func:`nxt.locator.find`.

    The :class:`Brick` object implements the context manager interface, so you can use
    it with the `with` syntax to close the connection when done with it.
    """

    def __init__(self, sock):
        self._sock = sock
        self._lock = threading.Lock()

    def play_tone_and_wait(self, frequency_hz, duration_ms):
        """Play a tone and wait until finished.

        :param int frequency_hz: Tone frequency in Hertz.
        :param int duration_ms: Tone duration in milliseconds.
        """
        self.play_tone(frequency_hz, duration_ms)
        time.sleep(duration_ms / 1000.0)

    def close(self):
        """Disconnect from the NXT brick."""
        if self._sock is not None:
            self._sock.close()
            self._sock = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def open_file(
        self,
        name,
        mode="r",
        size=None,
        *,
        buffering=-1,
        encoding=None,
        errors=None,
        newline=None
    ):
        """Open a file and return a corresponding file-like object.

        :param str name: Name of the file to open.
        :param str mode: Specification of open mode.
        :param int size: For writing, give the final size of the file.
        :param int buffering: Buffering control.
        :param encoding: Encoding for text mode.
        :type encoding: str or None
        :param errors: Encoding error handling for text mode.
        :type errors: str or None
        :param newline: Newline handling for text mode.
        :type newline: str or None
        :return: A file-like object connected to the file on the NXT brick.
        :rtype: io.IOBase
        :raises nxt.error.FileNotFoundError: When file does not exists.
        :raises nxt.error.FileExistsError: When file already exists.
        :raises nxt.error.SystemProtocolError: When no space is available.

        `mode` is a string which specifies how the file should be open. You can combine
        several characters to build the specification:

        =========  =====================================
        Character  Meaning
        =========  =====================================
        'r'        open for reading (default)
        'w'        open for writing (`size` must be given)
        't'        use text mode (default)
        'b'        use binary mode
        =========  =====================================

        When writing a file, the NXT brick needs to know the total size when opening the
        file, so this must be given as parameter.

        Other parameters (`buffering`, `encoding`, `errors` and `newline`) have the same
        meaning as the standard :func:`open` function, they must be given as keyword
        parameters.

        When `encoding` is None or not given, it defaults to ``ascii`` as this is the
        only encoding understood by the NXT brick.
        """
        rw = None
        tb = None
        for c in mode:
            if c in "rw" and rw is None:
                rw = c
            elif c in "tb" and tb is None:
                tb = c
            else:
                raise ValueError("invalid mode")
        if rw is None:
            raise ValueError("must give read or write mode")
        if tb is None:
            tb = "t"
        if tb == "b":
            if encoding is not None:
                raise ValueError("invalid encoding argument for binary mode")
            if errors is not None:
                raise ValueError("invalid errors argument for binary mode")
            if newline is not None:
                raise ValueError("invalid newline argument for binary mode")
        else:
            if buffering == 0:
                raise ValueError("invalid buffering argument for text mode")
            if encoding is None:
                encoding = "ascii"
        if buffering == -1:
            buffering = self._sock.bsize
        if rw == "r":
            if size is not None:
                raise ValueError("size given for reading")
            raw = RawFileReader(self, name)
            if buffering == 0:
                return raw
            buf = io.BufferedReader(raw, buffering)
        else:
            if size is None:
                raise ValueError("size not given for writing")
            raw = RawFileWriter(self, name, size)
            if buffering == 0:
                return raw
            buf = io.BufferedWriter(raw, buffering)
        if tb == "t":
            return io.TextIOWrapper(buf, encoding, errors, newline, buffering == 1)
        else:
            return buf

    def find_files(self, pattern="*.*"):
        """Find all files matching a pattern.

        :param str pattern: Pattern to match files against.
        :return: An iterator on all matching files, returning file name and
            file size as a tuple.
        :rtype: Iterator[str, int]

        Accepted patterns are:

        - ``*.*``: to match anything (default),
        - ``<name>.*``: to match files with any extension,
        - ``*.<extension>``: to match files with given extension,
        - ``<name>.<extension>``: to match using full name.
        """
        try:
            handle, name, size = self.file_find_first(pattern)
        except nxt.error.FileNotFoundError:
            return None
        try:
            yield name, size
            while True:
                try:
                    _, name, size = self.file_find_next(handle)
                except nxt.error.FileNotFoundError:
                    break
                yield name, size
        finally:
            self.file_close(handle)

    def find_modules(self, pattern="*.*"):
        """Find all modules matching a pattern.

        :param str pattern: Pattern to match modules against, use ``*.*``
            (default) to match any module.
        :return: An iterator on all matching modules, returning module name,
            identifier, size and IO map size as a tuple.
        :rtype: Iterator[str, int, int, int]
        """
        try:
            handle, mname, mid, msize, miomap_size = self.module_find_first(pattern)
        except nxt.error.ModuleNotFoundError:
            return None
        try:
            yield mname, mid, msize, miomap_size
            while True:
                try:
                    _, mname, mid, msize, miomap_size = self.module_find_next(handle)
                except nxt.error.ModuleNotFoundError:
                    break
                yield mname, mid, msize, miomap_size
        finally:
            self.module_close(handle)

    def get_motor(self, port):
        """Return a motor object connected to one of the brick output port.

        :param nxt.motor.Port port: Output port identifier.
        :return: The motor object.
        :rtype: nxt.motor.Motor
        """
        return nxt.motor.Motor(self, port)

    def get_sensor(self, port, cls=None, *args, **kwargs):
        """Return a sensor object connected to one of the brick input port.

        :param nxt.sensor.Port port: Input port identifier.
        :param cls: Sensor class, or None to autodetect.
        :type cls: typing.Type[nxt.sensor.Sensor] or None
        :param args: Additional constructor positional arguments when `cls` is given.
        :param kwargs: Additional constructor keyword arguments when `cls` is given.
        :return: A sensor object.
        :rtype: nxt.sensor.Sensor
        :raises nxt.sensor.digital.SearchError: When sensor can not be identified.

        When `cls` is not given or ``None``, try to detect the sensor type and return
        the correct sensor object. This only works for digital sensors with
        identification information.

        For autodetection to work, the module containing the sensor class must be
        imported at least once. See modules in :mod:`nxt.sensor`.
        """
        if cls is None:
            if args or kwargs:
                raise ValueError("extra arguments with autodetect")
            base_sensor = nxt.sensor.digital.BaseDigitalSensor(
                self, port, check_compatible=False
            )
            info = base_sensor.get_sensor_info()
            return nxt.sensor.digital.find_class(info)(
                self, port, check_compatible=False
            )
        else:
            return cls(self, port, *args, **kwargs)

    def _cmd(self, tgram):
        """Send a message to the NXT brick and read reply.

        :param nxt.telegram.Telegram tgram: Message to send.
        :return: Reply message after status has been checked, or None if no reply
           requested.
        :rtype: nxt.telegram.Telegram or None
        """
        reply_tgram = None
        with self._lock:
            self._sock.send(tgram.bytes())
            if tgram.reply_req:
                reply_tgram = Telegram(opcode=tgram.opcode, pkt=self._sock.recv())
        if reply_tgram:
            reply_tgram.check_status()
        return reply_tgram

    def start_program(self, name):
        """Start a program on the brick.

        :param str name: Program file name (example: ``"myprogram.rxe"``).
        """
        tgram = Telegram(Opcode.DIRECT_START_PROGRAM)
        tgram.add_filename(name)
        self._cmd(tgram)

    def stop_program(self):
        """Stop the running program on the brick.

        :raises nxt.error.DirectProtocolError: When no program is running.
        """
        tgram = Telegram(Opcode.DIRECT_STOP_PROGRAM)
        self._cmd(tgram)

    def play_sound_file(self, loop, name):
        """Play a sound file on the brick.

        :param bool loop: Loop mode, play continuously.
        :param str name: Sound file name.
        """
        tgram = Telegram(Opcode.DIRECT_PLAY_SOUND_FILE, reply_req=False)
        tgram.add_bool(loop)
        tgram.add_filename(name)
        self._cmd(tgram)

    def play_tone(self, frequency_hz, duration_ms):
        """Play a tone on the brick, do not wait until finished.

        :param int frequency_hz: Tone frequency in Hertz.
        :param int duration_ms: Tone duration in milliseconds.

        This function do not wait until finished, if you want to play several notes, you
        may need :func:`play_tone_and_wait`.
        """
        tgram = Telegram(Opcode.DIRECT_PLAY_TONE, reply_req=False)
        tgram.add_u16(frequency_hz)
        tgram.add_u16(duration_ms)
        self._cmd(tgram)

    def set_output_state(
        self, port, power, mode, regulation_mode, turn_ratio, run_state, tacho_limit
    ):
        """Set output port state on the brick.

        :param nxt.motor.Port port: Output port identifier.
        :param int power: Motor speed or power level (-100 to 100).
        :param nxt.motor.Mode mode: Motor power mode.
        :param nxt.motor.RegulationMode regulation_mode: Motor regulation mode.
        :param int turn_ratio: Turn ratio (-100 to 100). Negative value shift power to
           the left motor.
        :param nxt.motor.RunState run_state: Motor run state.
        :param int tacho_limit: Number of degrees the motor should rotate relative to
           the current position.

        .. warning:: This is a low level function, prefer to use
           :meth:`nxt.motor.Motor`, you can get one from :meth:`get_motor`.
        """
        tgram = Telegram(Opcode.DIRECT_SET_OUT_STATE, reply_req=False)
        tgram.add_u8(port.value)
        tgram.add_s8(power)
        tgram.add_u8(mode.value)
        tgram.add_u8(regulation_mode.value)
        tgram.add_s8(turn_ratio)
        tgram.add_u8(run_state.value)
        tgram.add_u32(tacho_limit)
        self._cmd(tgram)

    def set_input_mode(self, port, sensor_type, sensor_mode):
        """Set input port mode on the brick.

        :param nxt.sensor.Port port: Input port identifier.
        :param nxt.sensor.Type sensor_type: Sensor type.
        :param nxt.sensor.Mode sensor_mode: Sensor mode.

        .. warning:: This is a low level function, prefer to use a :mod:`nxt.sensor`
           class.
        """
        tgram = Telegram(Opcode.DIRECT_SET_IN_MODE, reply_req=False)
        tgram.add_u8(port.value)
        tgram.add_u8(sensor_type.value)
        tgram.add_u8(sensor_mode.value)
        self._cmd(tgram)

    def get_output_state(self, port):
        """Get output port state from the brick.

        :param nxt.motor.Port port: Output port identifier.
        :return: A tuple with `port`, `power`, `mode`, `regulation_mode`, `turn_ratio`,
           `run_state`, `tacho_limit`, `tacho_count`, `block_tacho_count`, and
           `rotation_count`.
        :rtype: (nxt.motor.Port, int, nxt.motor.Mode, nxt.motor.RegulationMode, int,
           nxt.motor.RunState, int, int, int, int)

        Return value details:

        - **port** Output port identifier.
        - **power** Motor speed or power level (-100 to 100).
        - **mode** Motor power mode.
        - **regulation_mode** Motor regulation mode.
        - **turn_ratio** Turn ratio (-100 to 100). Negative value shift power to the
          left motor.
        - **run_state** Motor run state.
        - **tacho_limit** Number of degrees the motor should rotate.
        - **block_tacho_count** Number of degrees the motor rotated relative to the
          "block" start.
        - **rotation_count** Number of degrees the motor rotated relative to the program
          start.

        .. warning:: This is a low level function, prefer to use
           :meth:`nxt.motor.Motor`, you can get one from :meth:`get_motor`.
        """
        tgram = Telegram(Opcode.DIRECT_GET_OUT_STATE)
        tgram.add_u8(port.value)
        tgram = self._cmd(tgram)
        port = nxt.motor.Port(tgram.parse_u8())
        power = tgram.parse_s8()
        mode = nxt.motor.Mode(tgram.parse_u8())
        regulation_mode = nxt.motor.RegulationMode(tgram.parse_u8())
        turn_ratio = tgram.parse_s8()
        run_state = nxt.motor.RunState(tgram.parse_u8())
        tacho_limit = tgram.parse_u32()
        tacho_count = tgram.parse_s32()
        block_tacho_count = tgram.parse_s32()
        rotation_count = tgram.parse_s32()
        return (
            port,
            power,
            mode,
            regulation_mode,
            turn_ratio,
            run_state,
            tacho_limit,
            tacho_count,
            block_tacho_count,
            rotation_count,
        )

    def get_input_values(self, port):
        """Get input port values from the brick.

        :param nxt.sensor.Port port: Input port identifier.
        :return: A tuple with `port`, `valid`, `calibrated`, `sensor_type`,
           `sensor_mode`, `raw_value`, `normalized_value`, `scaled_value`, and
           `calibrated_value`. `rotation_count`.
        :rtype: (nxt.sensor.Port, bool, bool, nxt.sensor.Type, nxt.sensor.Mode, int,
           int, int, int)

        Return value details:

        - **port** Input port identifier.
        - **valid** ``True`` if the value is valid, else ``False``.
        - **calibrated** Always ``False``, there is no calibration in NXT firmware.
        - **sensor_type** Sensor type.
        - **sensor_mode** Sensor mode.
        - **raw_value** Raw analog to digital converter value.
        - **normalized_value** Normalized value.
        - **scaled_value** Scaled value.
        - **calibrated_value** Always normalized value, there is no calibration in NXT
          firmware.

        .. warning:: This is a low level function, prefer to use a :mod:`nxt.sensor`
           class.
        """
        tgram = Telegram(Opcode.DIRECT_GET_IN_VALS)
        tgram.add_u8(port.value)
        tgram = self._cmd(tgram)
        port = nxt.sensor.Port(tgram.parse_u8())
        valid = tgram.parse_bool()
        calibrated = tgram.parse_bool()
        sensor_type = nxt.sensor.Type(tgram.parse_u8())
        sensor_mode = nxt.sensor.Mode(tgram.parse_u8())
        raw_value = tgram.parse_u16()
        normalized_value = tgram.parse_u16()
        scaled_value = tgram.parse_s16()
        calibrated_value = tgram.parse_s16()
        return (
            port,
            valid,
            calibrated,
            sensor_type,
            sensor_mode,
            raw_value,
            normalized_value,
            scaled_value,
            calibrated_value,
        )

    def reset_input_scaled_value(self, port):
        """Reset scaled value for an input port on the brick.

        :param nxt.sensor.Port port: Input port identifier.

        This can be used to reset accumulated value for some sensor modes.

        .. warning:: This is a low level function, prefer to use a :mod:`nxt.sensor`
           class.
        """
        tgram = Telegram(Opcode.DIRECT_RESET_IN_VAL)
        tgram.add_u8(port.value)
        self._cmd(tgram)

    def message_write(self, inbox, message):
        """Send a message to a brick mailbox.

        :param int inbox: Mailbox number (0 to 19).
        :param bytes message: Message to send (58 bytes maximum).
        """
        if len(message) > 58:
            raise ValueError("message too long")
        tgram = Telegram(Opcode.DIRECT_MESSAGE_WRITE)
        tgram.add_u8(inbox)
        tgram.add_u8(len(message) + 1)
        tgram.add_bytes(message)
        tgram.add_u8(0)
        self._cmd(tgram)

    def reset_motor_position(self, port, relative):
        """Reset block or program motor position for a brick output port.

        :param nxt.motor.Port port: Output port identifier.
        :param bool relative: If ``True``, reset block position, if ``False``, reset
           program position.

        .. warning:: This is a low level function, prefer to use
           :meth:`nxt.motor.Motor`, you can get one from :meth:`get_motor`.
        """
        tgram = Telegram(Opcode.DIRECT_RESET_POSITION)
        tgram.add_u8(port.value)
        tgram.add_bool(relative)
        self._cmd(tgram)

    def get_battery_level(self):
        """Get brick battery voltage.

        :return: Battery voltage in millivolt.
        :rtype: int
        """
        tgram = Telegram(Opcode.DIRECT_GET_BATT_LVL)
        tgram = self._cmd(tgram)
        millivolts = tgram.parse_u16()
        return millivolts

    def stop_sound_playback(self):
        """Stop currently running sound file on the brick."""
        tgram = Telegram(Opcode.DIRECT_STOP_SOUND)
        self._cmd(tgram)

    def keep_alive(self):
        """Reset the brick standby timer.

        :return: Sleep timeout in milliseconds.
        :rtype: int
        """
        tgram = Telegram(Opcode.DIRECT_KEEP_ALIVE)
        tgram = self._cmd(tgram)
        sleep_timeout = tgram.parse_u32()
        return sleep_timeout

    def ls_get_status(self, port):
        """Get status of last low-speed transaction to a brick input port.

        :param nxt.sensor.Port port: Input port identifier.
        :return: Number of bytes to read as a result of the transaction.
        :rtype: int
        :raises nxt.error.I2CPendingError: When transaction is still in progress.
        :raises nxt.error.DirectProtocolError: When there is an error on the bus.

        .. warning:: This is a low level function, prefer to use a :mod:`nxt.sensor`
           class.
        """
        tgram = Telegram(Opcode.DIRECT_LS_GET_STATUS)
        tgram.add_u8(port.value)
        tgram = self._cmd(tgram)
        size = tgram.parse_u8()
        return size

    def ls_write(self, port, tx_data, rx_bytes):
        """Write data to a brick input port using low speed transaction.

        :param nxt.sensor.Port port: Input port identifier.
        :param bytes tx_data: Data to send.
        :param int rx_bytes: Number of bytes to receive.

        Function returns immediately. Transaction status can be retrieved using
        :meth:`ls_get_status` and result must be read using :meth:`ls_read`.

        .. warning:: This is a low level function, prefer to use a :mod:`nxt.sensor`
           class.
        """
        tgram = Telegram(Opcode.DIRECT_LS_WRITE)
        tgram.add_u8(port.value)
        tgram.add_u8(len(tx_data))
        tgram.add_u8(rx_bytes)
        tgram.add_bytes(tx_data)
        self._cmd(tgram)

    def ls_read(self, port):
        """Read result of low speed transaction.

        :param nxt.sensor.Port port: Input port identifier.
        :return: Data received.
        :rtype: bytes
        :raises nxt.error.I2CPendingError: When transaction is still in progress.
        :raises nxt.error.DirectProtocolError: When there is an error on the bus.

        The :meth:`ls_write` function must be called to initiate the transaction.

        .. warning:: This is a low level function, prefer to use a :mod:`nxt.sensor`
           class.
        """
        tgram = Telegram(Opcode.DIRECT_LS_READ)
        tgram.add_u8(port.value)
        tgram = self._cmd(tgram)
        size = tgram.parse_u8()
        rx_data = tgram.parse_bytes(size)
        return rx_data

    def get_current_program_name(self):
        """Return name of program currently running on the brick.

        :return: Program file name
        :rtype: str
        :raises nxt.error.DirectProtocolError: When no program is running.
        """
        tgram = Telegram(Opcode.DIRECT_GET_CURR_PROGRAM)
        tgram = self._cmd(tgram)
        name = tgram.parse_filename()
        return name

    def message_read(self, remote_inbox, local_inbox, remove):
        """Read a message from a brick mailbox.

        :param int remote_inbox: Mailbox number (0 to 19).
        :param int local_inbox: Local mailbox, not used by brick.
        :param bool remove: Whether to remove the message from the mailbox.
        :return: The read message.
        :rtype: bytes
        :raises nxt.error.DirectProtocolError: When mailbox is empty.
        """
        tgram = Telegram(Opcode.DIRECT_MESSAGE_READ)
        tgram.add_u8(remote_inbox)
        tgram.add_u8(local_inbox)
        tgram.add_bool(remove)
        tgram = self._cmd(tgram)
        local_inbox = tgram.parse_u8()
        size = tgram.parse_u8()
        message = tgram.parse_bytes(size)
        return local_inbox, message

    def file_open_read(self, name):
        """Open file for reading.

        :param str name: File name.
        :return: The file handle and the file size.
        :rtype: (int, int)
        :raises nxt.error.FileNotFoundError: When file does not exists.

        .. warning:: This is a low level function, prefer to use :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_OPENREAD)
        tgram.add_filename(name)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        size = tgram.parse_u32()
        return handle, size

    def file_open_write(self, name, size):
        """Open file for writing.

        :param str name: File name.
        :param int size: Final file size.
        :return: The file handle.
        :rtype: int
        :raises nxt.error.FileExistsError: When file already exists.
        :raises nxt.error.SystemProtocolError: When no space is available.

        .. warning:: This is a low level function, prefer to use :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_OPENWRITE)
        tgram.add_filename(name)
        tgram.add_u32(size)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        return handle

    def file_read(self, handle, size):
        """Read data from open file.

        :param int handle: Open file handle.
        :param int size: Number of bytes to read.
        :return: The file handle and the read data.
        :rtype: (int, bytes)

        .. warning:: This is a low level function, prefer to use :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_READ)
        tgram.add_u8(handle)
        tgram.add_u16(size)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        size = tgram.parse_u16()
        data = tgram.parse_bytes(size)
        return handle, data

    def file_write(self, handle, data):
        """Write data to open file.

        :param int handle: Open file handle.
        :param bytes data: Data to write.
        :return: The file handle and the number of bytes written.
        :rtype: (int, int)

        .. warning:: This is a low level function, prefer to use :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_WRITE)
        tgram.add_u8(handle)
        tgram.add_bytes(data)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        size = tgram.parse_u16()
        return handle, size

    def file_close(self, handle):
        """Close open file.

        :param int handle: Open file handle.
        :return: The closed file handle.
        :rtype: int

        .. warning:: This is a low level function, prefer to use :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_CLOSE)
        tgram.add_u8(handle)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        return handle

    def file_delete(self, name):
        """Delete a file on the brick.

        :param str name: File name.
        :return: The deleted file name.
        :rtype: str
        :raises nxt.error.FileNotFoundError: When file does not exists.
        """
        tgram = Telegram(Opcode.SYSTEM_DELETE)
        tgram.add_filename(name)
        tgram = self._cmd(tgram)
        name = tgram.parse_filename()
        return name

    def file_find_first(self, pattern):
        """Start finding files matching a pattern.

        :param str pattern: Pattern to match files against.
        :return: A handle for the search, first file found name and size.
        :rtype: (int, str, int)
        :raises nxt.error.FileNotFoundError: When no file is found.

        .. warning:: This is a low level function, prefer to use :meth:`find_files`.
        """
        tgram = Telegram(Opcode.SYSTEM_FINDFIRST)
        tgram.add_filename(pattern)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        name = tgram.parse_filename()
        size = tgram.parse_u32()
        return handle, name, size

    def file_find_next(self, handle):
        """Continue finding files.

        :param int handle: Handle open with :meth:`file_find_first`.
        :return: The handle, next file found name and size.
        :rtype: (int, str, int)
        :raises nxt.error.FileNotFoundError: When no more file is found.

        .. warning:: This is a low level function, prefer to use :meth:`find_files`.
        """
        tgram = Telegram(Opcode.SYSTEM_FINDNEXT)
        tgram.add_u8(handle)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        name = tgram.parse_filename()
        size = tgram.parse_u32()
        return handle, name, size

    def get_firmware_version(self):
        """Get firmware version information.

        :return: Protocol and firmware versions, as two tuples with major and minor for
           each version.
        :rtype: ((int, int), (int, int))
        """
        tgram = Telegram(Opcode.SYSTEM_VERSIONS)
        tgram = self._cmd(tgram)
        prot_minor = tgram.parse_u8()
        prot_major = tgram.parse_u8()
        prot_version = (prot_major, prot_minor)
        fw_minor = tgram.parse_u8()
        fw_major = tgram.parse_u8()
        fw_version = (fw_major, fw_minor)
        return prot_version, fw_version

    def file_open_write_linear(self, name, size):
        """Open file for writing, reserve a linear space.

        :param str name: File name.
        :param int size: Final file size.
        :return: The file handle.
        :rtype: int
        :raises nxt.error.FileExistsError: When file already exists.
        :raises nxt.error.SystemProtocolError: When no space is available.

        Linear space is required for programs, but the brick will automatically use
        linear mode with :meth:`file_open_write` when extension is ``.rxe``, ``.sys`` or
        ``.rtm``.

        .. warning:: This is a low level function, prefer to use :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_OPENWRITELINEAR)
        tgram.add_filename(name)
        tgram.add_u32(size)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        return handle

    def file_open_write_data(self, name, size):
        """Open file for writing, using data mode.

        :param str name: File name.
        :param int size: Maximum file size.
        :return: The file handle.
        :rtype: int
        :raises nxt.error.FileExistsError: When file already exists.
        :raises nxt.error.SystemProtocolError: When no space is available.

        A data file can be written in small chunks, and can grow later. It can be used
        for data logging.

        .. warning:: This is a low level function, however, there is no support yet for
           data files in :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_OPENWRITEDATA)
        tgram.add_filename(name)
        tgram.add_u32(size)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        return handle

    def file_open_append_data(self, name):
        """Open file for appending, using data mode.

        :param str name: File name.
        :return: The file handle and available size
        :rtype: (int, int)
        :raises nxt.error.FileNotFoundError: When file does not exists.
        :raises nxt.error.SystemProtocolError: When file is full or file is not a data
           file.

        The file must be a data file. The available size is the size which has not been
        written to last time the file was open.

        .. warning:: This is a low level function, however, there is no support yet for
           data files in :meth:`open_file`.
        """
        tgram = Telegram(Opcode.SYSTEM_OPENAPPENDDATA)
        tgram.add_filename(name)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        available_size = tgram.parse_u32()
        return handle, available_size

    def module_find_first(self, pattern):
        """Start finding modules matching a pattern.

        :param str pattern: Pattern to match modules against.
        :return: A handle for the search, first module found name, identifier, size and
           IO map size.
        :rtype: (int, str, int, int, int)
        :raises nxt.error.ModuleNotFoundError: When no module is found.

        .. warning:: This is a low level function, prefer to use :meth:`find_modules`.
        """
        tgram = Telegram(Opcode.SYSTEM_FINDFIRSTMODULE)
        tgram.add_filename(pattern)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        name = tgram.parse_filename()
        mod_id = tgram.parse_u32()
        mod_size = tgram.parse_u32()
        mod_iomap_size = tgram.parse_u16()
        return handle, name, mod_id, mod_size, mod_iomap_size

    def module_find_next(self, handle):
        """Continue finding modules.

        :param int handle: Handle open with :meth:`module_find_first`.
        :return: The handle, next module found name, identifier, size and IO map size.
        :rtype: (int, str, int, int, int)
        :raises nxt.error.ModuleNotFoundError: When no more module is found.

        .. warning:: This is a low level function, prefer to use :meth:`find_modules`.
        """
        tgram = Telegram(Opcode.SYSTEM_FINDNEXTMODULE)
        tgram.add_u8(handle)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        name = tgram.parse_filename()
        mod_id = tgram.parse_u32()
        mod_size = tgram.parse_u32()
        mod_iomap_size = tgram.parse_u16()
        return handle, name, mod_id, mod_size, mod_iomap_size

    def module_close(self, handle):
        """Close module search.

        :param int handle: Open module handle.
        :return: The closed module handle.
        :rtype: int

        .. warning:: This is a low level function, prefer to use :meth:`find_modules`.
        """
        tgram = Telegram(Opcode.SYSTEM_CLOSEMODHANDLE)
        tgram.add_u8(handle)
        tgram = self._cmd(tgram)
        handle = tgram.parse_u8()
        return handle

    def read_io_map(self, mod_id, offset, size):
        """Read module IO map on the brick.

        :param int mod_id: Module identifier.
        :param int offset: Offset in IO map.
        :param int size: Number of bytes to read.
        :return: Module identifier and read data.
        :rtype: (int, bytes)

        Module identifier can be found using :meth:`find_modules`. You need to know the
        structure of the module IO map. You can find it by reading the firmware source
        code.
        """
        tgram = Telegram(Opcode.SYSTEM_IOMAPREAD)
        tgram.add_u32(mod_id)
        tgram.add_u16(offset)
        tgram.add_u16(size)
        tgram = self._cmd(tgram)
        mod_id = tgram.parse_u32()
        size = tgram.parse_u16()
        data = tgram.parse_bytes(size)
        return mod_id, data

    def write_io_map(self, mod_id, offset, data):
        """Write module IO map on the brick.

        :param int mod_id: Module identifier.
        :param int offset: Offset in IO map.
        :param bytes data: Data to write.
        :return: Module identifier and written size.
        :rtype: (int, int)

        Module identifier can be found using :meth:`find_modules`. You need to know the
        structure of the module IO map. You can find it by reading the firmware source
        code.
        """
        tgram = Telegram(Opcode.SYSTEM_IOMAPWRITE)
        tgram.add_u32(mod_id)
        tgram.add_u16(offset)
        tgram.add_u16(len(data))
        tgram.add_bytes(data)
        tgram = self._cmd(tgram)
        mod_id = tgram.parse_u32()
        size = tgram.parse_u16()
        return mod_id, size

    def boot(self, *, sure=False):
        """Erase NXT brick firmware and go to SAM-BA boot mode.

        :param bool sure: Set to ``True`` if you are really sure. Must be a keyword
           parameter.
        :return: Brick response, should be ``"Yes\0"``.
        :rtype: bytes

        This only works on USB connection.

        .. danger:: This **erases the firmware** of the brick, you need to send a new
           firmware to use it. Sending a firmware is not supported by NXT-Python. You
           can use the original LEGO software or `libnxt`_ for example. Be sure to know
           what you are doing.

        .. _libnxt: https://git.ni.fr.eu.org/libnxt.git/
        """
        if not sure:
            raise ValueError("this is dangerous, please read documentation")
        tgram = Telegram(Opcode.SYSTEM_BOOTCMD)
        tgram.add_bytes(b"Let's dance: SAMBA\0")
        tgram = self._cmd(tgram)
        resp = tgram.parse_bytes()
        return resp

    def set_brick_name(self, name):
        """Set brick name.

        :param str name: New brick name.
        """
        tgram = Telegram(Opcode.SYSTEM_SETBRICKNAME)
        tgram.add_string(15, name)
        self._cmd(tgram)

    def get_device_info(self):
        """Get brick information.

        :return: The brick name, Bluetooth address, Bluetooth signal strengths and free
           user flash.
        :rtype: (str, str, (int, int, int, int), int)

        Bluetooth address uses this notation: ``00:16:53:xx:xx:xx``, where `xx` is the
        brick unique address part (`00:16:53` is the LEGO OUI used for the NXT bricks).
        """
        tgram = Telegram(Opcode.SYSTEM_DEVICEINFO)
        tgram = self._cmd(tgram)
        name = tgram.parse_string(15)
        a0 = tgram.parse_u8()
        a1 = tgram.parse_u8()
        a2 = tgram.parse_u8()
        a3 = tgram.parse_u8()
        a4 = tgram.parse_u8()
        a5 = tgram.parse_u8()
        # a6 is not used, should be zero.
        tgram.parse_u8()
        address = "%02X:%02X:%02X:%02X:%02X:%02X" % (a0, a1, a2, a3, a4, a5)
        signal_strengths = (
            tgram.parse_u8(),
            tgram.parse_u8(),
            tgram.parse_u8(),
            tgram.parse_u8(),
        )
        user_flash = tgram.parse_u32()
        return name, address, signal_strengths, user_flash

    def delete_user_flash(self):
        """Erase the brick user flash."""
        tgram = Telegram(Opcode.SYSTEM_DELETEUSERFLASH)
        self._cmd(tgram)

    def poll_command_length(self, buf_num):
        """Get number of bytes available in brick poll buffer.

        :param int buf_num: Buffer number, 0 for USB, 1 for high speed.
        :return: Buffer number and number of available bytes.
        :rtype: (int, int)
        """
        tgram = Telegram(Opcode.SYSTEM_POLLCMDLEN)
        tgram.add_u8(buf_num)
        tgram = self._cmd(tgram)
        buf_num = tgram.parse_u8()
        size = tgram.parse_u8()
        return buf_num, size

    def poll_command(self, buf_num, size):
        """Get bytes from brick poll buffer.

        :param int buf_num: Buffer number, 0 for USB, 1 for high speed.
        :param int size: Number of bytes to read.
        :return: Buffer number and read bytes.
        :rtype: (int, bytes)
        """
        tgram = Telegram(Opcode.SYSTEM_POLLCMD)
        tgram.add_u8(buf_num)
        tgram.add_u8(size)
        tgram = self._cmd(tgram)
        buf_num = tgram.parse_u8()
        size = tgram.parse_u8()
        command = tgram.parse_bytes(size)
        return buf_num, command

    def bluetooth_factory_reset(self):
        """Reset brick Bluetooth to factory settings.

        This only works on USB connection.
        """
        tgram = Telegram(Opcode.SYSTEM_BTFACTORYRESET)
        self._cmd(tgram)



# nxt.locator module -- Locate NXT bricks
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
The :mod:`.locator` module allows to detect connected NXT bricks and to create
corresponding :class:`~nxt.brick.Brick` objects.

The :func:`find` function is your main starting point to create a NXT-Python program.

If you want to make a command line tool, :func:`add_arguments` and
:func:`find_with_options` will make it easy to allow choosing a brick from the command
line.
"""
import configparser
import importlib
import logging
import os

__all__ = ["find", "add_arguments", "find_with_options", "BrickNotFoundError"]

logger = logging.getLogger(__name__)


class BrickNotFoundError(Exception):
    """Exception raised when searching for a NXT brick, but no brick can be found."""

    pass


def _get_default_backends(**filters):
    """Get default backends names.

    :param filters: Additional filter keywords or backends parameters, used to select
       additional backend based on some filter parameters.
    """
    backends = []
    if "filename" in filters:
        backends.append("devfile")
    if "server_host" in filters or "server_port" in filters:
        backends.append("socket")
    backends.extend(["usb", "bluetooth"])
    return backends


def _get_backends(backends):
    """Get backends objects.

    :param backends: Specify backends to use.
    :type backends: list[str or object]
    :return: An iterator on the backends object list.
    :rtype: Iterator[object]
    """
    for backend in backends:
        if isinstance(backend, str):
            if not backend.isidentifier():
                raise ValueError("invalid backend identifier")
            module = importlib.import_module(f"nxt.backend.{backend}")
            backend = module.get_backend()
        if backend is not None:
            yield backend


def _get_config(config="default", config_filenames=None):
    """Read configuration file and get requested configuration.

    :param config: Name of the configuration file section to use, or ``None`` to disable
       configuration reading.
    :type config: str or None
    :param config_filenames: Configuration file paths, or ``None`` for default.
    :type config_filenames: Iterable[str or bytes or os.PathLike] or None
    :return: Configuration section or None.
    :rtype: MutableMapping[str, str]
    """
    if config is None:
        return None
    if config_filenames is None:
        config_filenames = [
            ".nxt-python.conf",
            os.path.expanduser("~/.nxt-python.conf"),
        ]
    parser = configparser.ConfigParser()
    logger.debug("configuration files=%s", config_filenames)
    read_filenames = parser.read(config_filenames)
    logger.debug("configuration read from %s", read_filenames)
    if config not in parser:
        logger.debug("no section %s, using %s", config, configparser.DEFAULTSECT)
        return parser[configparser.DEFAULTSECT]
    return parser[config]


def find(
    *,
    find_all=False,
    backends=None,
    custom_match=None,
    config="default",
    config_filenames=None,
    name=None,
    host=None,
    **filters,
):
    """Find a NXT brick and return it.

    :param bool find_all: ``True`` to return an iterator over all bricks found.
    :param backends: Specify backends to use, use ``None`` for default.
    :type backends: Iterable[str or object] or None
    :param custom_match: Function to filter bricks found.
    :type custom_match: Callable or None
    :param config: Name of the configuration file section to use, or ``None`` to disable
       configuration reading.
    :type config: str or None
    :param config_filenames: Configuration file paths, or ``None`` for default.
    :type config_filenames: Iterable[str or bytes or os.PathLike] or None
    :param name: Brick name (example: ``"NXT"``).
    :type name: str or None
    :param host: Bluetooth address (example: ``"00:16:53:01:02:03"``).
    :type host: str or None
    :param filters: Additional filter keywords or backends parameters.
    :return: The found brick, or an iterator if `find_all` is ``True``
    :rtype: nxt.brick.Brick or Iterator[nxt.brick.Brick]
    :raises BrickNotFoundError: if no brick is found and `find_all` is ``False``.

    Use this function to find a NXT brick. You can pass arguments to match a specific
    brick, for example, this will return the brick with name "NXT":

    >>> import nxt.locator
    >>> b = nxt.locator.find(name="NXT")

    If there is more than one matching brick, the first one found will be returned. If
    no brick is found, :exc:`BrickNotFoundError` is raised.

    If you want to find all matching bricks, you can set the `find_all` parameter to
    ``True``. It will return an iterator on all found bricks. If no brick is found, an
    empty iterator is returned.

    You can also use a custom function to search for your brick:

    >>> def is_my_brick(brick):
    ...     name = brick.get_device_info()[0]
    ...     return name.startswith("NXT")
    ...
    >>> for b in nxt.locator.find(find_all=True, custom_match=is_my_brick):
    ...     b.play_tone(440, 1000)
    ...

    The `name` and `host` parameters are passed to the backends, and are also used to
    filter the result, so you can use the `host` parameter even when not using
    Bluetooth.

    Extra keywords arguments are given to the backends which can use them or not. See
    the :mod:`nxt.backend` documentation.

    The `backends` parameter allows overriding the default list of backends to use. Each
    element of the list is a backend object or a backend name. Again, see
    :mod:`nxt.backend` documentation for the list of available backends.

    Configuration is used to load default values for `backends` parameter and selection
    parameters. If the `config` parameter is not ``None``, a configuration will be read
    from files listed by the `config_filenames` parameter, or from a default list of
    files. The `config` parameter corresponds to the section to use for configuration.
    """
    config_section = _get_config(config, config_filenames)

    if config_section is not None:
        if backends is None:
            backends = config_section.get("backends", None)
            if backends is not None:
                backends = backends.split()
        if custom_match is None and name is None and host is None and not filters:
            name = config_section.get("name", None)
            host = config_section.get("host", None)
            for key, value in config_section.items():
                if key not in ("backends", "name", "host"):
                    filters[key] = value

    if backends is None:
        backends = _get_default_backends(**filters)

    def iter_bricks():
        for backend in _get_backends(backends):
            logger.info("using backend from %s", backend.__module__)
            for brick in backend.find(name=name, host=host, **filters):
                logger.debug("found brick %s", brick)
                if name is not None or host is not None:
                    bname, bhost, _, _ = brick.get_device_info()
                    logger.debug("found brick with name=%s and host=%s", bname, bhost)
                    if name is not None and name != bname:
                        logger.debug("brick name mismatch, %s != %s", bname, name)
                        brick.close()
                        continue
                    if host is not None and host != bhost:
                        logger.debug("brick host mismatch, %s != %s", bhost, host)
                        brick.close()
                        continue
                if custom_match is not None and not custom_match(brick):
                    logger.debug("brick rejected by custom_match")
                    brick.close()
                    continue
                yield brick

    if find_all:
        return iter_bricks()
    else:
        brick = next(iter_bricks(), None)
        if brick is None:
            raise BrickNotFoundError("no brick found")
        return brick


def add_arguments(parser):
    """Add options to an :mod:`argparse` parser to allow configuration from the command
    line.

    :param argparse.ArgumentParser parser: An :mod:`argparse` parser.

    This can be used to easily design a command line interface. Use it with
    :func:`find_with_options`.

    Example:

    >>> import argparse
    >>> import nxt.locator
    >>> p = argparse.ArgumentParser(description="My NXT-Python program")
    >>> nxt.locator.add_arguments(p)
    >>> p.add_argument("--hello", help="say hello (example)")
    >>> options = p.parse_args()
    >>> brick = nxt.locator.find_with_options(options)
    """
    parser.add_argument(
        "--backend",
        dest="backends",
        action="append",
        choices=("usb", "bluetooth", "socket", "devfile"),
        help="enable backend, can be given several times",
    )
    parser.add_argument("--config", help="name of configuration file section to use")
    parser.add_argument(
        "--config-filename",
        dest="config_filenames",
        action="append",
        help="configuration life path, can be given several times",
    )
    parser.add_argument("--name", help="NXT brick name (example: NXT)")
    parser.add_argument(
        "--host", help="NXT brick Bluetooth address (example: 00:16:53:01:02:03)"
    )
    parser.add_argument(
        "--server-host", help="server address or name (example: localhost)"
    )
    parser.add_argument("--server-port", type=int, help="server port (example: 2727)")
    parser.add_argument("--filename", help="device file name (example: /dev/rfcomm0)")


def find_with_options(options, *, find_all=False):
    """Find a NXT brick and return it, using options from command line.

    :param argparse.Namespace options: Options returned by
       :meth:`argparse.ArgumentParser.parse_args`
    :param bool find_all: ``True`` to return an iterator over all bricks found.
    :return: The found brick or ``None``, or an iterator if `find_all` is ``True``.
    :rtype: nxt.brick.Brick or None or Iterator[nxt.brick.Brick]

    This is to be used together with :func:`add_arguments`. It calls :func:`find` with
    options received on the command line.
    """
    kwargs = dict()
    for k in (
        "backends",
        "config",
        "config_filenames",
        "name",
        "host",
        "server_host",
        "server_port",
        "filename",
    ):
        v = getattr(options, k)
        if v is not None:
            kwargs[k] = v
    return find(find_all=find_all, **kwargs)



# nxt.sensor.hitechnic module -- Classes to read HiTechnic sensors
# Copyright (C) 2006,2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, Paulo Vieira, rhn
# Copyright (C) 2010  rhn, Marcus Wanner, melducky, Samuel Leeman-Munk
# Copyright (C) 2011  jerradgenson, Marcus Wanner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from nxt.sensor import Type, Mode
from .digital import BaseDigitalSensor
from .analog import BaseAnalogSensor


class Compass(BaseDigitalSensor):
    """Hitechnic compass sensor."""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'mode': (0x41, 'B'),
                        'heading': (0x42, 'B'),
                        'adder' : (0x43, 'B'),
    })
    
    class Modes:
        MEASUREMENT = 0x00
        CALIBRATION = 0x43
        CALIBRATION_FAILED = 0x02
    
    def get_heading(self):
        """Returns heading from North in degrees."""

        two_degree_heading = self.read_value('heading')[0]
        adder = self.read_value('adder')[0]
        heading = two_degree_heading * 2 + adder

        return heading
    
    get_sample = get_heading

    def get_relative_heading(self,target=0):
        rheading = self.get_sample()-target
        if rheading > 180:
            rheading -= 360
        elif rheading < -180:
            rheading += 360
        return rheading	
    
    def is_in_range(self,minval,maxval):
        """This deserves a little explanation:
if max > min, it's straightforward, but
if min > max, it switches the values of max and min
and returns true if heading is NOT between the new max and min
        """
        if minval > maxval:
            (maxval,minval) = (minval,maxval)
            inverted = True
        else:
            inverted = False
        heading = self.get_sample()
        in_range = (heading > minval) and (heading < maxval)
        #an xor handles the reversal
        #a faster, more compact way of saying
        #if !reversed return in_range
        #if reversed return !in_range
        return bool(inverted) ^ bool(in_range) 

    def get_mode(self):
        return self.read_value('mode')[0]
    
    def set_mode(self, mode):
         if mode != self.Modes.MEASUREMENT and \
                 mode != self.Modes.CALIBRATION:
             raise ValueError('Invalid mode specified: ' + str(mode))
         self.write_value('mode', (mode, ))
         
Compass.add_compatible_sensor(None, 'HiTechnc', 'Compass ') #Tested with version '\xfdV1.23  '
Compass.add_compatible_sensor(None, 'HITECHNC', 'Compass ') #Tested with version '\xfdV2.1   '


class Accelerometer(BaseDigitalSensor):
    'Object for Accelerometer sensors. Thanks to Paulo Vieira.'
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'x_axis_high': (0x42, 'b'),
        'y_axis_high': (0x43, 'b'),
        'z_axis_high': (0x44, 'b'),
        'xyz_short': (0x42, '3b'),
        'all_data': (0x42, '3b3B')
    })
    
    class Acceleration:
        def __init__(self, x, y, z):
            self.x, self.y, self.z = x, y, z
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def get_acceleration(self):
        """Returns the acceleration along x, y, z axes. 200 => 1g.
        """
        xh, yh, zh, xl, yl, zl = self.read_value('all_data')
        x = xh << 2 | xl
        y = yh << 2 | yl
        z = zh << 2 | zl
        return self.Acceleration(x, y, z)
    
    get_sample = get_acceleration

Accelerometer.add_compatible_sensor(None, 'HiTechnc', 'Accel.  ')
Accelerometer.add_compatible_sensor(None, 'HITECHNC', 'Accel.  ') #Tested with version '\xfdV1.1   '


class IRReceiver(BaseDigitalSensor):
    """Object for HiTechnic IRReceiver sensors for use with LEGO Power Functions IR
Remotes. Coded to HiTechnic's specs for the sensor but not tested. Please report
whether this worked for you or not!
    """
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({
        'm1A': (0x42, 'b'),
        'm1B': (0x43, 'b'),
        'm2A': (0x44, 'b'),
        'm2B': (0x45, 'b'),
        'm3A': (0x46, 'b'),
        'm3B': (0x47, 'b'),
        'm4A': (0x48, 'b'),
        'm4B': (0x49, 'b'),
        'all_data': (0x42, '8b')
    })

    class SpeedReading:
        def __init__(self, m1A, m1B, m2A, m2B, m3A, m3B, m4A, m4B):
            self.m1A, self.m1B, self.m2A, self.m2B, self.m3A, self.m3B, self.m4A, self.m4B = m1A, m1B, m2A, m2B, m3A, m3B, m4A, m4B
            self.channel_1 = (m1A, m1B)
            self.channel_2 = (m2A, m2B)
            self.channel_3 = (m3A, m3B)
            self.channel_4 = (m4A, m4B)
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def get_speeds(self):
        """Returns the motor speeds for motors A and B on channels 1-4.
Values are -128, -100, -86, -72, -58, -44, -30, -16, 0, 16, 30, 44, 58, 72, 86
and 100. -128 specifies motor brake mode. Note that no motors are actually
being controlled here!
        """
        m1A, m1B, m2A, m2B, m3A, m3B, m4A, m4B = self.read_value('all_data')
        return self.SpeedReading(m1A, m1B, m2A, m2B, m3A, m3B, m4A, m4B)
    
    get_sample = get_speeds

IRReceiver.add_compatible_sensor(None, 'HiTechnc', 'IRRecv  ')
IRReceiver.add_compatible_sensor(None, 'HITECHNC', 'IRRecv  ')


class IRSeekerv2(BaseDigitalSensor):
    """Object for HiTechnic IRSeeker sensors. Coded to HiTechnic's specs for the sensor
but not tested. Please report whether this worked for you or not!
    """
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({
        'dspmode': (0x41, 'B'),
        'DC_direction': (0x42, 'B'),
        'DC_sensor_1': (0x43, 'B'),
        'DC_sensor_2': (0x44, 'B'),
        'DC_sensor_3': (0x45, 'B'),
        'DC_sensor_4': (0x46, 'B'),
        'DC_sensor_5': (0x47, 'B'),
        'DC_sensor_mean': (0x48, 'B'),
        'all_DC': (0x42, '7B'),
        'AC_direction': (0x49, 'B'),
        'AC_sensor_1': (0x4A, 'B'),
        'AC_sensor_2': (0x4B, 'B'),
        'AC_sensor_3': (0x4C, 'B'),
        'AC_sensor_4': (0x4D, 'B'),
        'AC_sensor_5': (0x4E, 'B'),
        'all_AC': (0x49, '6B')
    })
    I2C_DEV = 0x10 #different from standard 0x02
    
    class DSPModes:
        #Modes for modulated (AC) data.
        AC_DSP_1200Hz = 0x00
        AC_DSP_600Hz = 0x01
    
    class _data:
        def get_dir_brightness(self, direction):
            "Gets the brightness of a given direction (1-9)."
            if direction%2 == 1: #if it's an odd number
                val = getattr(self, "sensor_%d" % ((direction-1)//2+1))
            else:
                val = (getattr(self, f"sensor_{direction // 2}")
                       + getattr(self, f"sensor_{direction // 2 + 1}")) / 2
            return val
    
    class DCData(_data):
        def __init__(self, direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5, sensor_mean):
            self.direction, self.sensor_1, self.sensor_2, self.sensor_3, self.sensor_4, self.sensor_5, self.sensor_mean = direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5, sensor_mean
    
    class ACData(_data):
        def __init__(self, direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5):
            self.direction, self.sensor_1, self.sensor_2, self.sensor_3, self.sensor_4, self.sensor_5 = direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5
            
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def get_dc_values(self):
        """Returns the unmodulated (DC) values.
        """
        direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5, sensor_mean = self.read_value('all_DC')
        return self.DCData(direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5, sensor_mean)
    
    def get_ac_values(self):
        """Returns the modulated (AC) values. 600Hz and 1200Hz modes can be selected
between by using the set_dsp_mode() function.
        """
        direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5 = self.read_value('all_AC')
        return self.ACData(direction, sensor_1, sensor_2, sensor_3, sensor_4, sensor_5)
    
    def get_dsp_mode(self):
        return self.read_value('dspmode')[0]
    
    def set_dsp_mode(self, mode):
        self.write_value('dspmode', (mode, ))
    
    get_sample = get_ac_values

IRSeekerv2.add_compatible_sensor(None, 'HiTechnc', 'NewIRDir')
IRSeekerv2.add_compatible_sensor(None, 'HITECHNC', 'NewIRDir')


class EOPD(BaseAnalogSensor):
    """Object for HiTechnic Electro-Optical Proximity Detection sensors.
    """
    
    # To be divided by processed value.
    _SCALE_CONSTANT = 250

    # Maximum distance the sensor can detect.
    _MAX_DISTANCE = 1023

    def __init__(self, brick, port):
        super().__init__(brick, port)
        from math import sqrt
        self.sqrt = sqrt

    def set_range_long(self):
        ''' Choose this mode to increase the sensitivity
            of the EOPD sensor by approximately 4x. May
            cause sensor overload.
            '''

        self.set_input_mode(Type.LIGHT_ACTIVE, Mode.RAW)

    def set_range_short(self):
        ''' Choose this mode to prevent the EOPD sensor from
            being overloaded by white objects.
           '''

        self.set_input_mode(Type.LIGHT_INACTIVE, Mode.RAW)
    
    def get_raw_value(self):
        '''Unscaled value read from sensor.'''

        return self._MAX_DISTANCE - self.get_input_values().raw_value
    
    def get_processed_value(self):
        '''Derived from the square root of the raw value.'''

        return self.sqrt(self.get_raw_value())

    def get_scaled_value(self):
        ''' Returns a value that will scale linearly as distance
            from target changes. This is the method that should
            generally be called to get EOPD sensor data.
            '''

        try:
            result = self._SCALE_CONSTANT / self.get_processed_value()
            return result

        except ZeroDivisionError:
            return self._SCALE_CONSTANT
    
    get_sample = get_scaled_value


class Colorv2(BaseDigitalSensor):
    """Object for HiTechnic Color v2 Sensors. Coded to HiTechnic's specs for the sensor
but not tested. Please report whether this worked for you or not!"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({
        'mode': (0x41, 'B'),
        'number': (0x42, 'B'),
        'red': (0x43, 'B'),
        'green': (0x44, 'B'),
        'blue': (0x45, 'B'),
        'white': (0x46, 'B'),
        'index': (0x47, 'B'),
        'normred': (0x48, 'B'),
        'normgreen': (0x49, 'B'),
        'normblue': (0x4A, 'B'),
        'all_data': (0x42, '9B'),
        'rawred': (0x42, '<H'),
        'rawgreen': (0x44, '<H'),
        'rawblue': (0x46, '<H'),
        'rawwhite': (0x48, '<H'),
        'all_raw_data': (0x42, '<4H')
    })
    
    class Modes:
        ACTIVE = 0x00 #get measurements using get_active_color
        PASSIVE = 0x01 #get measurements using get_passive_color
        RAW = 0x03 #get measurements using get_passive_color
        BLACK_CALIBRATION = 0x42 #hold away from objects, results saved in EEPROM
        WHITE_CALIBRATION = 0x43 #hold in front of white surface, results saved in EEPROM
        LED_POWER_LOW = 0x4C #saved in EEPROM, must calibrate after using
        LED_POWER_HIGH = 0x48 #saved in EEPROM, must calibrate after using
        RANGE_NEAR = 0x4E #saved in EEPROM, only affects active mode
        RANGE_FAR = 0x46 #saved in EEPROM, only affects active mode, more susceptable to noise
        FREQ_50 = 0x35 #saved in EEPROM, use when local wall power is 50Hz
        FREQ_60 = 0x36 #saved in EEPROM, use when local wall power is 60Hz

    class ActiveData:
        def __init__(self, number, red, green, blue, white, index, normred, normgreen, normblue):
            self.number, self.red, self.green, self.blue, self.white, self.index, self.normred, self.normgreen, self.normblue = number, red, green, blue, white, index, normred, normgreen, normblue

    class PassiveData:
        #also holds raw mode data
        def __init__(self, red, green, blue, white):
            self.red, self.green, self.blue, self.white = red, green, blue, white
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def get_active_color(self):
        """Returns color values when in active mode.
        """
        number, red, green, blue, white, index, normred, normgreen, normblue = self.read_value('all_data')
        return self.ActiveData(number, red, green, blue, white, index, normred, normgreen, normblue)
    
    get_sample = get_active_color
    
    def get_passive_color(self):
        """Returns color values when in passive or raw mode.
        """
        red, green, blue, white = self.read_value('all_raw_data')
        return self.PassiveData(red, green, blue, white)
    
    def get_mode(self):
        return self.read_value('mode')[0]
    
    def set_mode(self, mode):
        self.write_value('mode', (mode, ))

Colorv2.add_compatible_sensor(None, 'HiTechnc', 'ColorPD')
Colorv2.add_compatible_sensor(None, 'HITECHNC', 'ColorPD')
Colorv2.add_compatible_sensor(None, 'HiTechnc', 'ColorPD ')
Colorv2.add_compatible_sensor(None, 'HITECHNC', 'ColorPD ')


class Gyro(BaseAnalogSensor):
    'Object for gyro sensors'
#This class is for the hitechnic gryo sensor. When the gryo is not
#moving there will be a constant offset that will change with 
#temperature and other ambient factors. The calibrate() function
#takes the currect value and uses it to offset subsequesnt ones.

    def __init__(self, brick, port):
        super().__init__(brick, port)
        self.set_input_mode(Type.ANGLE, Mode.RAW)
        self.offset = 0
    
    def get_rotation_speed(self):
        return self.get_input_values().scaled_value - self.offset
    
    def set_zero(self, value):
        self.offset = value
    
    def calibrate(self):
        self.set_zero(self.get_rotation_speed())
    
    get_sample = get_rotation_speed


class Prototype(BaseDigitalSensor):
    """Object for HiTechnic sensor prototype boards. Coded to HiTechnic's specs but not
tested. Please report whether this worked for you or not!
    """
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({
        'A0': (0x42, '<H'),
        'A0': (0x44, '<H'),
        'A0': (0x46, '<H'),
        'A0': (0x48, '<H'),
        'A0': (0x4A, '<H'),
        'all_analog': (0x42, '<5H'),
        'digital_in': (0x4C, 'B'),
        'digital_out': (0x4D, 'B'),
        'digital_cont': (0x4E, 'B'),
        'sample_time': (0x4F, 'B'),
        })
    
    class Digital_Data():
        """Container for 6 bits of digital data. Takes an integer or a list of six bools
and can be converted into a list of bools or an integer."""
        def __init__(self, pins):
            if isinstance(pins, int):
                self.dataint = pins
                self.datalst = self.tolist(pins)
            else:
                self.dataint = self.toint(pins)
                self.datalst = pins
            self.d0, self.d1, self.d2, self.d3, self.d4, self.d5 = self.datalst
        
        def tolist(self, val):
            lst = []
            for i in range(6):
                lst.append(bool(val & 2**i))
            return lst

        def toint(self, lst):
            val = 0
            for i in range(6):
                val += int(bool(lst[i])) * (2**i)
            return val
        
        def __int__(self):
            return self.dataint
        
        def __iter__(self):
            return iter(self.datalst)
        
        def __getitem__(self, i):
            return self.datalst[i]
    
    class Analog_Data():
        def __init__(self, a0, a1, a2, a3, a4):
            self.a0, self.a1, self.a2, self.a3, self.a4 = a0, a1, a2, a3, a4
    
    def get_analog(self):
        return self.Analog_Data(*self.read_value('all_analog'))
    
    def get_digital(self):
        return self.Digital_Data(self.read_value('digital_in')[0])
    
    def set_digital(self, pins):
        """Can take a Digital_Data() object"""
        self.write_value('digital_out', (int(pins), ))
    
    def set_digital_modes(self, modes):
        """Sets input/output mode of digital pins. Can take a Digital_Data() object."""
        self.write_value('digital_cont', (int(modes), ))

Prototype.add_compatible_sensor(None, 'HiTechnc', 'Proto   ')


class ServoCon(BaseDigitalSensor):
    """Object for HiTechnic FIRST Servo Controllers. Coded to HiTechnic's specs for
the sensor but not tested. Please report whether this worked for you or not!"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({
        'status': (0x40, 'B'),
        'steptime': (0x41, 'B'),
        's1pos': (0x42, 'B'),
        's2pos': (0x43, 'B'),
        's3pos': (0x44, 'B'),
        's4pos': (0x45, 'B'),
        's5pos': (0x46, 'B'),
        's6pos': (0x47, 'B'),
        'pwm': (0x48, 'B'),
    })
    
    class Status:
        STOPPED = 0x00 #all motors stopped
        RUNNING = 0x01 #motor(s) moving

    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def get_status(self):
        """Returns the status of the motors. 0 for all stopped, 1 for
some running.
        """
        return self.read_value('status')[0]
    
    def set_step_time(self, time):
        """Sets the step time (0-15).
        """
        self.write_value('steptime', (time, ))
    
    def set_pos(self, num, pos):
        """Sets the position of a server. num is the servo number (1-6),
pos is the position (0-255).
        """
        self.write_value('s%dpos' % num, (pos, ))
    
    def get_pwm(self):
        """Gets the "PWM enable" value. The function of this value is
nontrivial and can be found in the documentation for the sensor.
        """
        return self.read_value('pwm')[0]
    
    def set_pwm(self, pwm):
        """Sets the "PWM enable" value. The function of this value is
nontrivial and can be found in the documentation for the sensor.
        """
        self.write_value('pwm', (pwm, ))

ServoCon.add_compatible_sensor(None, 'HiTechnc', 'ServoCon')


class MotorCon(BaseDigitalSensor):
    """Object for HiTechnic FIRST Motor Controllers."""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({
        'm1enctarget': (0x40, '>l'),
        'm1mode': (0x44, 'B'),
        'm1power': (0x45, 'b'),
        'm2power': (0x46, 'b'),
        'm2mode': (0x47, 'B'),
        'm2enctarget': (0x48, '>l'),
        'm1enccurrent': (0x4c, '>l'),
        'm2enccurrent': (0x50, '>l'),
        'batteryvoltage': (0x54, '2B'),
        'm1gearratio': (0x56, 'b'),
        'm1pid': (0x57, '3B'),
        'm2gearratio': (0x5a, 'b'),
        'm2pid': (0x5b, '3B'),
    })
    
    class PID_Data():
        def __init__(self, p, i, d):
            self.p, self.i, self.d = p, i, d
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)
    
    def set_enc_target(self, mot, val):
        """Set the encoder target (-2147483648-2147483647) for a motor
        """
        self.write_value('m%denctarget'%mot, (val, ))
    
    def get_enc_target(self, mot):
        """Get the encoder target for a motor
        """
        return self.read_value('m%denctarget'%mot)[0]
    
    def get_enc_current(self, mot):
        """Get the current encoder value for a motor
        """
        return self.read_value('m%denccurrent'%mot)[0]
    
    def set_mode(self, mot, mode):
        """Set the mode for a motor. This value is a bit mask and you can
find details about it in the sensor's documentation.
        """
        self.write_value('m%dmode'%mot, (mode, ))
    
    def get_mode(self, mot):
        """Get the mode for a motor. This value is a bit mask and you can
find details about it in the sensor's documentation.
        """
        return self.read_value('m%dmode'%mot)[0]
    
    def set_power(self, mot, power):
        """Set the power (-100-100) for a motor
        """
        self.write_value('m%dpower'%mot, (power, ))
    
    def get_power(self, mot):
        """Get the power for a motor
        """
        return self.read_value('m%dpower'%mot)[0]
    
    def set_gear_ratio(self, mot, ratio):
        """Set the gear ratio for a motor
        """
        self.write_value('m%dgearratio'%mot, (ratio, ))
    
    def get_gear_ratio(self, mot):
        """Get the gear ratio for a motor
        """
        return self.read_value('m%dgearratio'%mot)[0]
    
    def set_pid(self, mot, piddata):
        """Set the PID coefficients for a motor. Takes data in
MotorCon.PID_Data(p, i, d) format.
        """
        self.write_value('m%dpid'%mot, (piddata.p, piddata.i, piddata.d))
    
    def get_pid(self, mot):
        """Get the PID coefficients for a motor. Returns a PID_Data() object.
        """
        p, i, d = self.read_value('m%dpid'%mot)
        return self.PID_Data(p, i, d)
    
    def get_battery_voltage(self):
        """Gets the battery voltage (in millivolts/20)
        """
        data = self.read_value('batteryvoltage')
        high = data[0]
        low = data[1]
        return high << 2 | low

MotorCon.add_compatible_sensor(None, 'HiTechnc', 'MotorCon')


class Angle(BaseDigitalSensor):
    """HiTechnic Angle Sensor."""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({
        'mode': (0x41, 'c'),
        'angle': (0x42, '2B'),
        'angle_acc': (0x44, '>l'),
        'rpm': (0x48, '>h'),
    })

    def get_angle(self):
        v = self.read_value('angle')
        return v[0] * 2 + v[1]

    get_sample = get_angle

    def get_accumulated_angle(self):
        return self.read_value("angle_acc")[0]

    def get_rpm(self):
        return self.read_value("rpm")[0]

    def calibrate(self):  #Current angle will be zero degrees written in EEPROM
        self.write_value('mode', b'C')

    def reset(self):    #Reset accumulated angle
        self.write_value('mode', b'R')



# conftest -- Common test fixtures
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import Mock, patch

import pytest

import nxt.brick


def pytest_addoption(parser):
    parser.addoption(
        "--run-nxt",
        action="append",
        choices=("usb", "bluetooth", "devfile", "socket"),
        default=[],
        help="run tests needing a real NXT connected over given interface",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "nxt: mark test as needing a real NXT")


def pytest_collection_modifyitems(config, items):
    run_nxt = config.getoption("--run-nxt")
    for item in items:
        for marker in item.iter_markers("nxt"):
            if marker.args[0] not in run_nxt:
                item.add_marker(
                    pytest.mark.skip(
                        reason="need --run-nxt=%s option to run" % marker.args[0]
                    )
                )


@pytest.fixture
def mtime():
    """Mock time.time() and time.sleep()."""
    current_time = 0

    def timef():
        nonlocal current_time
        return current_time

    def sleepf(delay):
        nonlocal current_time
        current_time += delay

    mtime = Mock(spec_set=("time", "sleep"))
    mtime.time.side_effect = timef
    mtime.sleep.side_effect = sleepf

    with patch("nxt.brick.time", new=mtime), patch(
        "nxt.motcont.time", new=mtime
    ), patch("nxt.motor.time", new=mtime), patch("nxt.sensor.digital.time", new=mtime):
        yield mtime


def make_brick_mock():
    b = Mock()

    def find_files(pattern):
        return nxt.brick.Brick.find_files(b, pattern)

    def find_modules(pattern):
        return nxt.brick.Brick.find_modules(b, pattern)

    def open_file(*args, **kwargs):
        return nxt.brick.Brick.open_file(b, *args, **kwargs)

    def get_sensor(*args, **kwargs):
        return nxt.brick.Brick.get_sensor(b, *args, **kwargs)

    def get_motor(*args, **kwargs):
        return nxt.brick.Brick.get_motor(b, *args, **kwargs)

    b._sock.bsize = 60
    b._sock.type = "usb"
    b.find_files = find_files
    b.find_modules = find_modules
    b.open_file = open_file
    b.get_sensor = get_sensor
    b.get_motor = get_motor
    return b


@pytest.fixture
def mbrick(mtime):
    """A brick with mocked low level functions."""
    return make_brick_mock()


@pytest.fixture
def mbrick2(mtime):
    """A second brick with mocked low level functions."""
    return make_brick_mock()



# nxt.backend.bluetooth module -- Bluetooth backend
# Copyright (C) 2006, 2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import logging
import struct

import nxt.brick

logger = logging.getLogger(__name__)

# NXT brick RFCOMM port.
PORT = 1


class BluetoothSock:
    """Bluetooth socket connected to a NXT brick."""

    #: Block size.
    bsize = 118

    #: Connection type, used to evaluate latency.
    type = "bluetooth"

    def __init__(self, bluetooth, host):
        self._bluetooth = bluetooth
        self._host = host
        self._sock = None

    def __str__(self):
        return f"Bluetooth ({self._host})"

    def connect(self):
        """Connect to NXT brick.

        :return: Connected brick.
        :rtype: Brick
        """
        logger.info("connecting via %s", self)
        sock = self._bluetooth.BluetoothSocket(self._bluetooth.RFCOMM)
        sock.connect((self._host, PORT))
        self._sock = sock
        return nxt.brick.Brick(self)

    def close(self):
        """Close the connection."""
        if self._sock is not None:
            logger.info("closing %s connection", self)
            self._sock.close()
            self._sock = None

    def send(self, data):
        """Send raw data.

        :param bytes data: Data to send.
        """
        data = struct.pack("<H", len(data)) + data
        logger.debug("send: %s", data.hex())
        self._sock.send(data)

    def recv(self):
        """Receive raw data.

        :return: Received data.
        :rtype: bytes
        """
        data = self._sock.recv(2)
        logger.debug("recv: %s", data.hex())
        (plen,) = struct.unpack("<H", data)
        data = self._sock.recv(plen)
        logger.debug("recv: %s", data.hex())
        return data


class Backend:
    """Bluetooth backend."""

    def __init__(self, bluetooth):
        self._bluetooth = bluetooth

    def find(self, host=None, name=None, **kwargs):
        """Find bricks connected using Bluetooth.

        :param host: Bluetooth address (example: ``"00:16:53:01:02:03"``).
        :type host: str or None
        :param name: Brick name (example: ``"NXT"``).
        :type name: str or None
        :param kwargs: Other parameters are ignored.
        :return: Iterator over all found bricks.
        :rtype: Iterator[Brick]
        """
        if host is not None:
            name = None
            lookup_names = False
            discovered = [host]
        else:
            lookup_names = name is not None
            try:
                discovered = self._bluetooth.discover_devices(lookup_names=lookup_names)
            except self._bluetooth.BluetoothError as err:
                logger.warning("failed to discover Bluetooth devices: %s", err)
                logger.debug("error from discover_devices", exc_info=True)
                return None
            except OSError as err:
                logger.warning("failed to discover Bluetooth devices: %s", err)
                logger.debug("error from discover_devices", exc_info=True)
                return None
        for dev in discovered:
            if lookup_names:
                devhost, devname = dev
            else:
                devhost, devname = dev, None
            if (host is None or devhost == host) and (name is None or devname == name):
                sock = BluetoothSock(self._bluetooth, devhost)
                try:
                    brick = sock.connect()
                except self._bluetooth.BluetoothError:
                    logger.warning("failed to connect to device %s", sock)
                    logger.debug("error from connect", exc_info=True)
                else:
                    yield brick


def get_backend():
    """Get an instance of the Bluetooth backend if available.

    :return: Bluetooth backend.
    :rtype: Backend or None
    """
    try:
        import bluetooth
    except ImportError:
        logger.info("no bluetooth module")
        return None
    except Exception:
        logger.info("platform is not supported by bluetooth module")
        logger.debug("error from import", exc_info=True)
        return None
    return Backend(bluetooth)



# nxt.backend.usb module -- USB backend
# Copyright (C) 2006, 2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner
# Copyright (C) 2011  Paul Hollensen, Marcus Wanner
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import logging
import os

import usb.core

import nxt.brick

logger = logging.getLogger(__name__)

# LEGO USB vendor ID.
ID_VENDOR_LEGO = 0x0694

# NXT brick product ID.
ID_PRODUCT_NXT = 0x0002


class USBSock:
    """USB socket connected to a NXT brick."""

    #: Block size.
    bsize = 60

    #: Connection type, used to evaluate latency.
    type = "usb"

    def __init__(self, dev):
        self._dev = dev
        self._epout = None
        self._epin = None

    def __str__(self):
        return "USB (Bus %03d Device %03d)" % (self._dev.bus, self._dev.address)

    def connect(self):
        """Connect to NXT brick.

        :return: Connected brick.
        :rtype: Brick
        """
        logger.info("connecting via %s", self)
        if os.name != "nt":
            # Do not reset device on Windows, see
            # https://github.com/schodet/nxt-python/issues/182 and
            # https://github.com/schodet/nxt-python/issues/33
            self._dev.reset()
        self._dev.set_configuration()
        intf = self._dev.get_active_configuration()[(0, 0)]
        self._epout, self._epin = intf
        return nxt.brick.Brick(self)

    def close(self):
        """Close the connection."""
        if self._epout is not None:
            logger.info("closing %s connection", self)
            self._epout = None
            self._epin = None

    def send(self, data):
        """Send raw data.

        :param bytes data: Data to send.
        """
        logger.debug("send: %s", data.hex())
        self._epout.write(data)

    def recv(self):
        """Receive raw data.

        :return: Received data.
        :rtype: bytes
        """
        data = self._epin.read(64).tobytes()
        logger.debug("recv: %s", data.hex())
        return data


class Backend:
    """USB backend."""

    def find(self, **kwargs):
        """Find bricks connected using USB.

        :param kwargs: Parameters are ignored.
        :return: Iterator over all found bricks.
        :rtype: Iterator[Brick]
        """
        for dev in usb.core.find(
            find_all=True, idVendor=ID_VENDOR_LEGO, idProduct=ID_PRODUCT_NXT
        ):
            sock = USBSock(dev)
            brick = sock.connect()
            yield brick


def get_backend():
    """Get an instance of the Bluetooth backend.

    :return: Bluetooth backend.
    :rtype: Backend
    """
    return Backend()



# nxt.sensor module -- Classes to read LEGO Mindstorms NXT sensors
# Copyright (C) 2006,2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, Paulo Vieira, rhn
# Copyright (C) 2010,2011,2012  Marcus Wanner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import logging
import struct
import time

from nxt.error import I2CError, I2CPendingError, DirectProtocolError
import nxt.sensor

logger = logging.getLogger(__name__)


class SensorInfo:
    def __init__(self, version, product_id, sensor_type):
        self.version = version
        self.product_id = product_id
        self.sensor_type = sensor_type

    def clarifybinary(self, instr, label):
        outstr = ''
        outstr += (label + ': `' + instr + '`\n')
        for char in instr:
            outstr += (hex(ord(char))+', ')
        outstr += ('\n')
        return outstr

    def __str__(self):
        outstr = ''
        outstr += (self.clarifybinary(str(self.version), 'Version'))
        outstr += (self.clarifybinary(str(self.product_id), 'Product ID'))
        outstr += (self.clarifybinary(str(self.sensor_type), 'Type'))
        return outstr

class BaseDigitalSensor(nxt.sensor.Sensor):
    """Object for digital sensors. I2C_ADDRESS is the dictionary storing name
    to i2c address mappings. It should be updated in every subclass. When
    subclassing this class, make sure to call add_compatible_sensor to add
    compatible sensor data.
    """
    I2C_DEV = 0x02
    I2C_ADDRESS = {'version': (0x00, '8s'),
        'product_id': (0x08, '8s'),
        'sensor_type': (0x10, '8s'),
#        'factory_zero': (0x11, 1),      # is this really correct?
        'factory_scale_factor': (0x12, 'B'),
        'factory_scale_divisor': (0x13, 'B'),
    }

    def __init__(self, brick, port, check_compatible=True):
        """Creates a BaseDigitalSensor. If check_compatible is True, queries
        the sensor for its name, and if a wrong sensor class was used, prints
        a warning.
        """
        super().__init__(brick, port)
        self.set_input_mode(nxt.sensor.Type.LOW_SPEED_9V, nxt.sensor.Mode.RAW)
        self.last_poll = time.time()
        self.poll_delay = 0.01
        time.sleep(0.1)  # Give I2C time to initialize
        #Don't do type checking if this class has no compatible sensors listed.
        try: self.compatible_sensors
        except AttributeError: check_compatible = False
        if check_compatible:
            sensor = self.get_sensor_info()
            if not sensor in self.compatible_sensors:
                logger.warning(
                    "wrong sensor class chosen for sensor %s on port %s",
                    sensor.product_id,
                    port,
                )
                logger.warning(
                    "   You may be using the wrong type of sensor or may have "
                    "connected the cable incorrectly. If you are sure you're using the "
                    "correct sensor class for the sensor, this message is likely in "
                    "error and you should disregard it and file a bug report, "
                    "including the output of get_sensor_info(). This message can be "
                    "suppressed by passing check_compatible=False when creating the "
                    "sensor object."
                )

    def _ls_get_status(self, size):
        for n in range(30): #https://code.google.com/p/nxt-python/issues/detail?id=35
            try:
                b = self._brick.ls_get_status(self._port)
                if b >= size:
                    return b
            except I2CPendingError:
                pass
        raise I2CError('ls_get_status timeout')

    def _i2c_command(self, address, value, format):
        """Writes an i2c value to the given address. value must be a string. value is
        a tuple of values corresponding to the given format.
        """
        value = struct.pack(format, *value)
        msg = bytes((self.I2C_DEV, address)) + value
        now = time.time()
        if self.last_poll+self.poll_delay > now:
            diff = now - self.last_poll
            time.sleep(self.poll_delay - diff)
        self.last_poll = time.time()
        self._brick.ls_write(self._port, msg, 0)

    def _i2c_query(self, address, format):
        """Reads an i2c value from given address, and returns a value unpacked
        according to the given format. Format is the same as in the struct
        module. See http://docs.python.org/library/struct.html#format-strings
        """
        size = struct.calcsize(format)
        msg = bytes((self.I2C_DEV, address))
        now = time.time()
        if self.last_poll+self.poll_delay > now:
            diff = now - self.last_poll
            time.sleep(self.poll_delay - diff)
        self.last_poll = time.time()
        self._brick.ls_write(self._port, msg, size)
        try:
            self._ls_get_status(size)
        finally:
            #we should clear the buffer no matter what happens
            data = self._brick.ls_read(self._port)
        if len(data) < size:
            raise I2CError('Read failure: Not enough bytes')
        data = struct.unpack(format, data[-size:])
        return data

    def read_value(self, name):
        """Reads a value from the sensor. Name must be a string found in
        self.I2C_ADDRESS dictionary. Entries in self.I2C_ADDRESS are in the
        name: (address, format) form, with format as in the struct module.
        Be careful on unpacking single variables - struct module puts them in
        tuples containing only one element.
        """
        address, fmt = self.I2C_ADDRESS[name]
        for n in range(3):
            try:
                return self._i2c_query(address, fmt)
            except DirectProtocolError:
                pass
        raise I2CError("read_value timeout")

    def write_value(self, name, value):
        """Writes value to the sensor. Name must be a string found in
        self.I2C_ADDRESS dictionary. Entries in self.I2C_ADDRESS are in the
        name: (address, format) form, with format as in the struct module.
        value is a tuple of values corresponding to the format from
        self.I2C_ADDRESS dictionary.
        """
        address, fmt = self.I2C_ADDRESS[name]
        self._i2c_command(address, value, fmt)

    def get_sensor_info(self):
        version = self.read_value('version')[0].decode('windows-1252').split('\0')[0]
        product_id = self.read_value('product_id')[0].decode('windows-1252').split('\0')[0]
        sensor_type = self.read_value('sensor_type')[0].decode('windows-1252').split('\0')[0]
        return SensorInfo(version, product_id, sensor_type)

    @classmethod
    def add_compatible_sensor(cls, version, product_id, sensor_type):
        """Adds an entry in the compatibility table for the sensor. If version
        is None, then it's the default class for this model. If product_id is
        None, then this is the default class for this vendor.
        """
        try:
            cls.compatible_sensors
        except AttributeError:
            cls.compatible_sensors = []
        finally:
            cls.compatible_sensors.append(SCompatibility(version, product_id,
                                                                            sensor_type))
            add_mapping(cls, version, product_id, sensor_type)


class SCompatibility(SensorInfo):
    """An object that helps manage the sensor mappings"""
    def __eq__(self, other):
        if self.product_id is None:
            return self.product_id == other.product_id
        elif self.version is None:
            return (self.product_id == other.product_id and
                    self.sensor_type == other.sensor_type)
        else:
            return (self.version == other.version and
                    self.product_id == other.product_id and
                    self.sensor_type == other.sensor_type)

sensor_mappings = {}


def add_mapping(cls, version, product_id, sensor_type):
    "None means any other value"
    if product_id not in sensor_mappings:
        sensor_mappings[product_id] = {}
    models = sensor_mappings[product_id]

    if sensor_type is None:
        if sensor_type in models:
            raise ValueError('Already registered!')
        models[sensor_type] = cls
        return

    if sensor_type not in models:
        models[sensor_type] = {}
    versions = models[sensor_type]

    if version in versions:
        raise ValueError('Already registered!')
    else:
        versions[version] = cls


class SearchError(Exception):
    pass


def find_class(info):
    """Returns an appropriate class for the given SensorInfo"""
    dic = sensor_mappings
    for val, msg in zip((info.product_id, info.sensor_type, info.version),
                                    ('Vendor', 'Model', 'Version')):
        if val in dic:
            dic = dic[val]
        elif None in dic:
            dic = dic[None]
        else:
            raise SearchError(msg + ' not found')
        return dic[info.sensor_type][None]



# test_backend_usb -- Test nxt.backend.usb module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
import array
from unittest.mock import Mock, call, patch

import pytest

import nxt.backend.usb


@pytest.fixture
def mdev():
    dev = Mock(
        spec_set=(
            "reset",
            "set_configuration",
            "get_active_configuration",
            "bus",
            "address",
        )
    )
    dev.bus = 1
    dev.address = 2
    return dev


@pytest.fixture
def musb(mdev):
    with patch("nxt.backend.usb.usb.core") as usb_core:
        usb_core.find.return_value = [mdev]
        yield usb_core


def test_usb(musb, mdev):
    # Instantiate backend.
    backend = nxt.backend.usb.get_backend()
    # Find brick.
    epout = Mock(spec_set=("write",))
    epin = Mock(spec_set=("read",))
    mdev.get_active_configuration.return_value = {(0, 0): (epout, epin)}
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 1
    brick = bricks[0]
    assert mdev.reset.called
    assert mdev.set_configuration.called
    assert mdev.get_active_configuration.called
    sock = brick._sock
    # str.
    assert str(sock).startswith("USB (Bus")
    # Send.
    some_bytes = bytes.fromhex("01020304")
    sock.send(some_bytes)
    assert epout.write.call_args == call(some_bytes)
    # Recv.
    epin.read.return_value = array.array("B", (1, 2, 3, 4))
    r = sock.recv()
    assert r == some_bytes
    assert epin.read.called
    # Close.
    brick.close()
    # Duplicated close.
    sock.close()


@pytest.mark.nxt("usb")
def test_usb_real():
    # Instantiate backend.
    backend = nxt.backend.usb.get_backend()
    # Find brick.
    bricks = list(backend.find())
    assert len(bricks) > 0, "no NXT found"
    brick = bricks[0]
    sock = brick._sock
    # str.
    assert str(sock).startswith("USB (Bus")
    # Send.
    sock.send(bytes.fromhex("019b"))
    # Recv.
    r = sock.recv()
    assert r.startswith(bytes.fromhex("029b00"))
    # Close.
    brick.close()



#!/usr/bin/python3
"""NXT-Python tutorial: increase log level."""
import logging

import nxt.locator

# Increase the log level, must be done before using any NXT-Python function. See logging
# documentation for details.
logging.basicConfig(level=logging.DEBUG)

with nxt.locator.find() as b:
    b.play_tone(440, 250)



# nxt.motcont module -- Interface to Linus Atorf's MotorControl NXC
# Copyright (C) 2011  Marcus Wanner
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import time
from threading import Lock

import nxt.error
import nxt.motor

__all__ = ["MotCont"]


def _power(power):
    pw = abs(power)
    if power < 0:
        pw += 100
    return "{:03}".format(pw)


def _tacho(tacholimit):
    return "{:06}".format(tacholimit)


class MotCont:
    """Interface to MotorControl.

    This class provides an interface to Linus Atorf's MotorControl NXC program. It is
    a wrapper which follows the `MotorControl documentation`_ and provides command
    strings and timing intervals as dictated there.

    To use this module, you will need to put ``MotorControl22.rxe`` on your NXT brick.
    It can be built using NXC and the corresponding source can be found at
    https://github.com/schodet/MotorControl.

    Use :func:`MotCont.start` to start the program. You can also start it manually my
    using the menu on the brick. When your script exits, it would be a good idea to do
    :func:`MotCont.stop`.

    Original `MotorControl site`_ is no longer available, but you can still find
    a mirror on web archive.

    .. _MotorControl documentation:
        https://github.com/schodet/MotorControl/blob/master/doc/MotorControl.md
    .. _MotorControl site:
        http://www.mindstorms.rwth-aachen.de/trac/wiki/MotorControl
    """

    def __init__(self, brick):
        self._brick = brick
        self._is_ready_lock = Lock()
        self._last_is_ready = time.time() - 1
        self._last_cmd = {}

    def _interval_is_ready(self):
        delay = 0.010
        diff = time.time() - self._last_is_ready
        if diff < delay:
            time.sleep(delay - diff)

    def _interval_motors(self, ports):
        delay = 0.015
        now = time.time()
        diff = delay
        for port in ports:
            if port in self._last_cmd:
                diff = min(diff, now - self._last_cmd[port])
        if diff < delay:
            time.sleep(delay - diff)

    def _record_time_motors(self, ports):
        now = time.time()
        for port in ports:
            self._last_cmd[port] = now

    def _decode_ports(self, ports, max_ports):
        try:
            ports = frozenset(ports)
        except TypeError:
            ports = frozenset((ports,))
        mapping = {
            frozenset((nxt.motor.Port.A,)): "0",
            frozenset((nxt.motor.Port.B,)): "1",
            frozenset((nxt.motor.Port.C,)): "2",
            frozenset((nxt.motor.Port.A, nxt.motor.Port.B)): "3",
            frozenset((nxt.motor.Port.A, nxt.motor.Port.C)): "4",
            frozenset((nxt.motor.Port.B, nxt.motor.Port.C)): "5",
            frozenset((nxt.motor.Port.A, nxt.motor.Port.B, nxt.motor.Port.C)): "6",
        }
        if ports not in mapping or len(ports) > max_ports:
            raise ValueError("invalid combination of ports")
        return ports, mapping[ports]

    def cmd(
        self, ports, power, tacholimit, speedreg=True, smoothstart=False, brake=False
    ):
        """Send a CONTROLLED_MOTORCMD to MotorControl.

        :param ports: Port or ports to control, use one of the port identifiers, or
           an iterable returning one or two of them.
        :type ports: nxt.motor.Port or Iterable[nxt.motor.Port]
        :param int power: Speed or power, -100 to 100.
        :param int tacholimit: Position to drive to, 1 to 999999, or 0 for unlimited.
        :param bool speedreg: ``True`` to enable regulation.
        :param bool smoothstart: ``True`` to enable smooth start.
        :param bool brake: ``True`` to hold brake at end of movement.
        """
        self._interval_is_ready()
        ports, strports = self._decode_ports(ports, 2)
        self._interval_motors(ports)
        mode = str(0x01 * int(brake) + 0x02 * int(speedreg) + 0x04 * int(smoothstart))
        command = "1" + strports + _power(power) + _tacho(tacholimit) + mode
        self._brick.message_write(1, command.encode("ascii"))
        self._record_time_motors(ports)

    def reset_tacho(self, ports):
        """Reset NXT tacho count.

        :param ports: Port or ports to control, use one of the port identifiers, or
           an iterable returning one to three of them.
        :type ports: nxt.motor.Port or Iterable[nxt.motor.Port]
        """
        self._interval_is_ready()
        ports, strports = self._decode_ports(ports, 3)
        command = "2" + strports
        self._brick.message_write(1, command.encode("ascii"))
        self._record_time_motors(ports)

    def is_ready(self, port):
        """Determine the state of a single motor.

        :param port: Port to control, use one of the port identifiers, or an iterable
           returning one of them.
        :type port: nxt.motor.Port or Iterable[nxt.motor.Port]
        :return: ``True`` if the motor is ready to accept new commands.
        :rtype: bool
        """
        self._interval_is_ready()
        ports, strports = self._decode_ports(port, 1)
        with self._is_ready_lock:
            command = "3" + strports
            self._brick.message_write(1, command.encode("ascii"))
            time.sleep(0.015)  # 10ms pause from the docs seems to not be adequate
            reply = self._brick.message_read(0, 1, 1)[1]
            if chr(reply[0]) != strports:
                raise nxt.error.ProtocolError("wrong port returned from ISMOTORREADY")
        self._last_is_ready = time.time()
        return bool(int(chr(reply[1])))

    def set_output_state(self, ports, power, tacholimit, speedreg=True):
        """Send a CLASSIC_MOTORCMD to MotorControl.

        :param ports: Port or ports to control, use one of the port identifiers, or
           an iterable returning one or two of them.
        :type ports: nxt.motor.Port or Iterable[nxt.motor.Port]
        :param int power: Speed or power, -100 to 100.
        :param int tacholimit: Position to drive to, 1 to 999999, or 0 for unlimited.
        :param bool speedreg: ``True`` to enable regulation.
        """
        self._interval_is_ready()
        ports, strports = self._decode_ports(ports, 2)
        self._interval_motors(ports)
        command = "4" + strports + _power(power) + _tacho(tacholimit) + str(speedreg)
        self._brick.message_write(1, command.encode("ascii"))
        self._record_time_motors(ports)

    def start(self, version=22):
        """Start the MotorControl program on the brick.

        :param int version: Version to start, default to 22 (version 2.2).

        It needs to already be present on the brick's flash and named
        ``MotorControlXX.rxe``, where `XX` is the version number passed as the version
        argument.
        """
        try:
            self._brick.stop_program()
            time.sleep(1)
        except nxt.error.DirectProtocolError:
            pass
        self._brick.start_program("MotorControl%d.rxe" % version)
        time.sleep(0.1)

    def stop(self):
        """Stop the MotorControl program.

        All this actually does is to stop the currently running program.
        """
        self._brick.stop_program()



# nxt.backend.socket module -- Socket backend
# Copyright (C) 2011  zonedaobne, Marcus Wanner
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import logging
import socket

import nxt.brick

logger = logging.getLogger(__name__)


class SocketSock:
    """Socket socket connected to a NXT brick."""

    #: Block size, conservative.
    bsize = 60

    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._sock = None
        #: Connection type, used to evaluate latency, known on connection.
        self.type = None

    def __str__(self):
        return f"Socket ({self._host}:{self._port})"

    def connect(self):
        """Connect to NXT brick.

        :return: Connected brick.
        :rtype: Brick
        """
        logger.info("connecting via %s:%d", self._host, self._port)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self._host, self._port))
        self._sock = sock
        self.send(bytes((0x98,)))
        self.type = "ip" + self.recv().decode("ascii")
        return nxt.brick.Brick(self)

    def close(self):
        """Close the connection."""
        if self._sock is not None:
            logger.info("closing connection to %s:%d", self._host, self._port)
            self._sock.send(bytes((0x99,)))
            self._sock.close()
            self._sock = None
            self.type = None

    def send(self, data):
        """Send raw data.

        :param bytes data: Data to send.
        """
        logger.debug("send: %s", data.hex())
        self._sock.send(data)

    def recv(self):
        """Receive raw data.

        :return: Received data.
        :rtype: bytes
        """
        data = self._sock.recv(1024)
        logger.debug("recv: %s", data.hex())
        return data


class Backend:
    """Socket backend.

    To be used with ``nxt_server`` script to access a NXT brick over the network.
    """

    def find(self, server_host="localhost", server_port=2727, **kwargs):
        """Find bricks connected using a socket.

        :param str server_host: Server address or name, default to `localhost`.
        :param server_port: Server port, default to 2727.
        :type server_port: str or int
        :param kwargs: Other parameters are ignored.
        :return: Iterator over all found bricks.
        :rtype: Iterator[Brick]
        """
        sock = SocketSock(server_host, int(server_port))
        try:
            brick = sock.connect()
        except ConnectionRefusedError:
            logger.exception("failed to connect to device %s", sock)
        else:
            yield brick


def get_backend():
    """Get an instance of the Socket backend.

    :return: Socket backend.
    :rtype: Backend
    """
    return Backend()



# nxt.motor module -- Class to control LEGO Mindstorms NXT motors
# Copyright (C) 2006  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, rhn
# Copyright (C) 2010  rhn
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import enum
import logging
import time

logger = logging.getLogger(__name__)


class Port(enum.Enum):
    """Output port identifier."""

    A = 0
    """Output port A."""

    B = 1
    """Output port B."""

    C = 2
    """Output port C."""


class Mode(enum.Flag):
    """Motor mode.

    These are flags that can be combined together using the `|` operator.
    """

    IDLE = 0x00
    """Keep the motor unpowered."""

    ON = 0x01
    """Enable motor power."""

    BRAKE = 0x02
    """Enable brake, motor input voltage is not left floating, must also be
    :attr:`ON`."""

    REGULATED = 0x04
    """Enable regulation, must also be :attr:`ON`."""


class RegulationMode(enum.Enum):
    """Motor regulation mode."""

    IDLE = 0
    """No regulation."""

    SPEED = 1
    """Speed regulation, :attr:`Mode.REGULATED` must be enabled."""

    SYNC = 2
    """Synchronous regulation of two motors, :attr:`Mode.REGULATED` must be enabled."""


class RunState(enum.Enum):
    """Motor run state."""

    IDLE = 0x00
    """Not running."""

    RAMP_UP = 0x10
    """Ramping speed up."""

    RUNNING = 0x20
    """Running at constant speed."""

    RAMP_DOWN = 0x40
    """Ramping speed down."""


LIMIT_RUN_FOREVER = 0
"""No angle limit."""


class BlockedException(Exception):
    """Raised when a motor is not moving as expected."""

    pass


class OutputState:
    """An object holding the internal state of a motor, not including rotation
    counters."""

    def __init__(self, values):
        (
            self.power,
            self.mode,
            self.regulation_mode,
            self.turn_ratio,
            self.run_state,
            self.tacho_limit,
        ) = values

    def to_list(self):
        """Returns a list of properties that can be used with set_output_state."""
        return [
            self.power,
            self.mode,
            self.regulation_mode,
            self.turn_ratio,
            self.run_state,
            self.tacho_limit,
        ]

    def __str__(self):
        return ", ".join(
            [
                str(self.mode),
                str(self.regulation_mode),
                str(self.turn_ratio),
                str(self.run_state),
                str(self.tacho_limit),
            ]
        )


class TachoInfo:
    """An object containing the information about the rotation of a motor."""

    def __init__(self, values):
        self.tacho_count, self.block_tacho_count, self.rotation_count = values

    def get_target(self, tacho_limit, direction):
        """Returns a TachoInfo object which corresponds to tacho state after
        moving for tacho_limit ticks in the given direction.

        :param int tacho_limit: Move angle.
        :param int direction: 1 (add) or -1 (subtract).
        :return: Updated state.
        :rtype: TachoInfo
        """
        # TODO: adjust other fields
        if abs(direction) != 1:
            raise ValueError("invalid direction")
        new_tacho = self.tacho_count + direction * tacho_limit
        return TachoInfo([new_tacho, None, None])

    def is_greater(self, target, direction):
        return direction * (self.tacho_count - target.tacho_count) > 0

    def is_near(self, target, threshold):
        difference = abs(target.tacho_count - self.tacho_count)
        return difference < threshold

    def __str__(self):
        return str((self.tacho_count, self.block_tacho_count, self.rotation_count))


class SynchronizedTacho:
    def __init__(self, leader_tacho, follower_tacho):
        self.leader_tacho = leader_tacho
        self.follower_tacho = follower_tacho

    def get_target(self, tacho_limit, direction):
        """This method will leave follower's target as None"""
        leader_tacho = self.leader_tacho.get_target(tacho_limit, direction)
        return SynchronizedTacho(leader_tacho, None)

    def is_greater(self, other, direction):
        return self.leader_tacho.is_greater(other.leader_tacho, direction)

    def is_near(self, other, threshold):
        return self.leader_tacho.is_near(other.leader_tacho, threshold)

    def __str__(self):
        if self.follower_tacho is not None:
            t2 = str(self.follower_tacho.tacho_count)
        else:
            t2 = "None"
        t1 = str(self.leader_tacho.tacho_count)
        return f"tacho: {t1} {t2}"


def get_tacho_and_state(values):
    """A convenience function. values is the list of values from
    get_output_state. Returns both OutputState and TachoInfo.
    """
    return OutputState(values[1:7]), TachoInfo(values[7:])


class BaseMotor:
    """Base class for motors"""

    def turn(self, power, tacho_units, brake=True, timeout=1, emulate=True):
        """Use this to turn a motor.

        :param int power: Value between -127 and 128 (an absolute value greater than 64
           is recommended)
        :param int tacho_units: Number of degrees to turn the motor. Values smaller than
           50 are not recommended and may have strange results.
        :param bool brake: Whether or not to hold the motor after the function exits
           (either by reaching the distance or throwing an exception).
        :param int timeout: Number of seconds after which a BlockedException is raised
           if the motor doesn't turn.
        :param bool emulate: If set to ``False``, the motor is aware of the tacho limit.
           If ``True``, a run() function equivalent is used. Warning: motors remember
           their positions and not using emulate may lead to strange behavior,
           especially with synced motors

        The motor will not stop until it turns the desired distance. Accuracy is much
        better over a USB connection than with bluetooth...
        """

        tacho_limit = tacho_units

        if tacho_limit < 0:
            raise ValueError("tacho_units must be greater than 0!")
        # TODO Calibrate the new values for ip socket latency.
        if self.method == "bluetooth":
            threshold = 70
        elif self.method == "usb":
            threshold = 5
        elif self.method == "ipbluetooth":
            threshold = 80
        elif self.method == "ipusb":
            threshold = 15
        else:
            threshold = 30  # compromise

        tacho = self.get_tacho()
        state = self._get_new_state()

        # Update modifiers even if they aren't used, might have been changed
        state.power = power
        if not emulate:
            state.tacho_limit = tacho_limit

        logger.debug("updating motor information")
        self._set_state(state)

        direction = 1 if power > 0 else -1
        logger.debug("tachocount: %s", tacho)
        current_time = time.time()
        tacho_target = tacho.get_target(tacho_limit, direction)

        blocked = False
        try:
            while True:
                time.sleep(self._eta(tacho, tacho_target, power) / 2)

                if not blocked:  # if still blocked, don't reset the counter
                    last_tacho = tacho
                    last_time = current_time

                tacho = self.get_tacho()
                current_time = time.time()
                blocked = self._is_blocked(tacho, last_tacho, direction)
                if blocked:
                    logger.debug("not advancing: %s %s", last_tacho, tacho)
                    # The motor can be up to 80+ degrees in either direction from target
                    # when using Bluetooth.
                    if current_time - last_time > timeout:
                        if tacho.is_near(tacho_target, threshold):
                            break
                        else:
                            raise BlockedException("Blocked!")
                else:
                    logger.debug("advancing: %s %s", last_tacho, tacho)
                if tacho.is_near(tacho_target, threshold) or tacho.is_greater(
                    tacho_target, direction
                ):
                    break
        finally:
            if brake:
                self.brake()
            else:
                self.idle()


class Motor(BaseMotor):
    def __init__(self, brick, port):
        self.brick = brick
        self.port = port
        self._read_state()
        self.sync = 0
        self.turn_ratio = 0
        self.method = brick._sock.type

    def _set_state(self, state):
        logger.debug("setting brick output state: %s", state)
        list_state = [self.port] + state.to_list()
        self.brick.set_output_state(*list_state)
        self._state = state

    def _read_state(self):
        logger.debug("getting brick output state")
        values = self.brick.get_output_state(self.port)
        self._state, tacho = get_tacho_and_state(values)
        return self._state, tacho

    # def get_tacho_and_state here would allow tacho manipulation

    def _get_state(self):
        """Returns a copy of the current motor state for manipulation."""
        return OutputState(self._state.to_list())

    def _get_new_state(self):
        state = self._get_state()
        if self.sync:
            state.mode = Mode.ON | Mode.REGULATED
            state.regulation_mode = RegulationMode.SYNC
            state.turn_ratio = self.turn_ratio
        else:
            state.mode = Mode.ON | Mode.REGULATED
            state.regulation_mode = RegulationMode.SPEED
        state.run_state = RunState.RUNNING
        state.tacho_limit = LIMIT_RUN_FOREVER
        return state

    def get_tacho(self):
        return self._read_state()[1]

    def reset_position(self, relative):
        """Resets the counters. Relative can be True or False"""
        self.brick.reset_motor_position(self.port, relative)

    def run(self, power=100, regulated=False):
        """Tells the motor to run continuously.

        :param int power: Motor power or speed if `regulated`.
        :param bool regulated: If ``True``, use speed regulation.
        """
        state = self._get_new_state()
        state.power = power
        if not regulated:
            state.mode = Mode.ON
        self._set_state(state)

    def brake(self):
        """Holds the motor in place"""
        state = self._get_new_state()
        state.power = 0
        state.mode = Mode.ON | Mode.BRAKE | Mode.REGULATED
        self._set_state(state)

    def idle(self):
        """Tells the motor to stop whatever it's doing. It also desyncs it."""
        state = self._get_new_state()
        state.power = 0
        state.mode = Mode.IDLE
        state.regulation_mode = RegulationMode.IDLE
        state.run_state = RunState.IDLE
        self._set_state(state)

    def weak_turn(self, power, tacho_units):
        """Tries to turn a motor for the specified distance. This function
        returns immediately, and it's not guaranteed that the motor turns that
        distance. This is an interface to use tacho_limit without
        RegulationMode.SPEED
        """
        tacho_limit = tacho_units
        state = self._get_new_state()

        # Update modifiers even if they aren't used, might have been changed
        state.mode = Mode.ON
        state.regulation_mode = RegulationMode.IDLE
        state.power = power
        state.tacho_limit = tacho_limit

        logger.debug("updating motor information")
        self._set_state(state)

    def _eta(self, current, target, power):
        """Returns time in seconds. Do not trust it too much"""
        tacho = abs(current.tacho_count - target.tacho_count)
        return (float(tacho) / abs(power)) / 5

    def _is_blocked(self, tacho, last_tacho, direction):
        """Returns if any of the engines is blocked"""
        return direction * (last_tacho.tacho_count - tacho.tacho_count) >= 0


class SynchronizedMotors(BaseMotor):
    """The object used to make two motors run in sync. Many objects may be
    present at the same time but they can't be used at the same time.
    Warning! Movement methods reset tacho counter.
    THIS CODE IS EXPERIMENTAL!!!
    """

    def __init__(self, leader, follower, turn_ratio):
        """Turn ratio can be >= 0 only! If you want to have it reversed,
        change motor order.
        """
        if follower.brick != leader.brick:
            raise ValueError("motors belong to different bricks")
        self.leader = leader
        self.follower = follower
        # Being from the same brick, they both have the same com method.
        self.method = self.leader.method

        if turn_ratio < 0:
            raise ValueError("turn ratio < 0, change motor order instead")

        if self.leader.port == self.follower.port:
            raise ValueError("The same motor passed twice")
        elif self.leader.port.value > self.follower.port.value:
            self.turn_ratio = turn_ratio
        else:
            logger.debug("reversed")
            self.turn_ratio = -turn_ratio

    def _get_new_state(self):
        return self.leader._get_new_state()

    def _set_state(self, state):
        self.leader._set_state(state)
        self.follower._set_state(state)

    def get_tacho(self):
        leadertacho = self.leader.get_tacho()
        followertacho = self.follower.get_tacho()
        return SynchronizedTacho(leadertacho, followertacho)

    def reset_position(self, relative):
        """Resets the counters. Relative can be True or False"""
        self.leader.reset_position(relative)
        self.follower.reset_position(relative)

    def _enable(self):  # This works as expected. I'm not sure why.
        # self._disable()
        self.reset_position(True)
        self.leader.sync = True
        self.follower.sync = True
        self.leader.turn_ratio = self.turn_ratio
        self.follower.turn_ratio = self.turn_ratio

    def _disable(self):  # This works as expected. (tacho is reset ok)
        self.leader.sync = False
        self.follower.sync = False
        # self.reset_position(True)
        self.leader.idle()
        self.follower.idle()
        # self.reset_position(True)

    def run(self, power=100):
        """Warning! After calling this method, make sure to call idle. The
        motors are reported to behave wildly otherwise.
        """
        self._enable()
        self.leader.run(power, True)
        self.follower.run(power, True)

    def brake(self):
        self._disable()  # reset the counters
        self._enable()
        self.leader.brake()  # brake both motors at the same time
        self.follower.brake()
        self._disable()  # now brake as usual
        self.leader.brake()
        self.follower.brake()

    def idle(self):
        self._disable()

    def turn(self, power, tacho_units, brake=True, timeout=1):
        self._enable()
        # non-emulation is a nightmare, tacho is being counted differently
        try:
            if power < 0:
                self.leader, self.follower = self.follower, self.leader
            BaseMotor.turn(self, power, tacho_units, brake, timeout, emulate=True)
        finally:
            if power < 0:
                self.leader, self.follower = self.follower, self.leader

    def _eta(self, tacho, target, power):
        return self.leader._eta(tacho.leader_tacho, target.leader_tacho, power)

    def _is_blocked(self, tacho, last_tacho, direction):
        # no need to check both, they're synced
        return self.leader._is_blocked(
            tacho.leader_tacho, last_tacho.leader_tacho, direction
        )
